import numpy as np
from meshparty import utils, skeleton
from copy import copy


def annotation_location_indices(mesh, anno_df, pos_column, sk_map=None, max_dist=np.inf,
                                voxel_resolution=np.array([4,4,40])):
    '''
    For a synapse dataframe associated with a given neuron, find the mesh indices associated with each synapse.

    :param mesh: trimesh Mesh
    :param synapse_df: DataFrame with at least one position column
    :param pos_column: string, column of dataframe to use for annotation positions
    :param sk_map: Optional, Numpy array with skeleton vertex index for every mesh vertex index.
    :param max_dist: Optional, Maximum distance to the mesh allowed for assignment, else return -1.
    :param voxel_resolution: Optional, default is [4,4,40] nm/voxel.
    :returns: Mesh indices and, if desired, skeleton indices.
    '''
    if len(anno_df) == 0:
        if sk_map is None:
            return np.array([])
        else:
            return np.array([]), np.array([])

    anno_positions = np.vstack(anno_df[pos_column].values) * voxel_resolution
    ds, mesh_inds = mesh.kdtree.query(anno_positions)
    mesh_inds[ds>max_dist] = -1
    
    if sk_map is None:
        return mesh_inds
    else:
        sk_map=sk_map.astype(int)
        found_inds = mesh_inds>=0
        skinds = np.zeros(mesh_inds.shape)
        skinds[found_inds] = sk_map[mesh_inds[found_inds]]
        skinds[~found_inds] = -1
        return mesh_inds, skinds


def annotation_skeleton_segments(sk, anno_df, pos_column, mesh=None, max_dist=np.inf,
                                 voxel_resolution=np.array([4,4,40]), skeleton_index_col_name=None):
    '''
    Attach skeleton segment index to an annotation dataframe
    :param sk: Skeleton
    :param anno_df: Annotation dataframe
    :param pos_column: String. Column name in dataframe with position values
    :param mesh: optional, mesh object. Needed if skeleton_index_col_name is not specified.
    :param max_dist: optional, float. Max distance to mesh for attaching annotations. Default is inf.
    :param voxel_resolution: optional, length 3 array. Default is [4,4,40] nm/pixel.
    :param skeleton_index_col_name: optional, string. Column name of skeleton vertex in dataframe.
    '''
    if mesh is None and anno_skind_col is None:
        raise ValueError('Must have either a mesh or existing skeleton indices')   

    if skeleton_index_col_name is None:
        sk_map = sk.mesh_to_skel_map
        minds, skinds = annotation_location_indices(mesh, anno_df, pos_column, sk_map=sk_map,
                                                max_dist=max_dist, voxel_resolution=voxel_resolution)
        anno_segments = sk.segment_map[skinds]
        return anno_segments, minds, skinds
    else:
        anno_segments = sk.segment_map[anno_df[skeleton_index_col_name]]
        return anno_segments


def skind_to_anno_map(sk, anno_df, pos_column=None, mesh=None, max_dist=np.inf,
                      voxel_resolution=np.array([4,4,40]), skeleton_index_col_name=None):
    '''
    Make a dict with key skeleton index and values a list of annotation ids at that index.
    '''
    anno_dict = {}
    if len(anno_df) == 0:
        return anno_dict

    if skeleton_index_col_name is None:
        minds, skinds = annotation_location_indices(mesh, anno_df, pos_column, sk_map=sk.mesh_to_skel_map,
                                                    max_dist=max_dist, voxel_resolution=voxel_resolution)
        anno_df = anno_df.copy()
        skeleton_index_col_name = 'XXX_temp_skeleton_index_internal'
        anno_df[skeleton_index_col_name] = skinds

    for k, v in anno_df[[skeleton_index_col_name, 'id']].groupby(skeleton_index_col_name).agg(lambda x: [int(y) for y in x]).to_dict()['id'].items():
        anno_dict[int(k)] = v
    return anno_dict


def synapse_betweenness(sk, pre_inds, post_inds):
    '''
    Compute synapse betweeness (number of paths from an input to an output) for all nodes of a skeleton.
    :param sk: Skeleton
    :param pre_inds: List of skeleton indices with an input synapse (Duplicate indices with more than one)
    :param post_inds: List of skeleton indices with an output synapse (Duplicate indices with more than one)

    :returns: Array of synapse betweenness for every vertex in the skeleton.
    '''
    def _precompute_synapse_inds(syn_inds):
        Nsyn = len(syn_inds)
        n_syn = np.zeros(len(sk.vertices))
        for ind in syn_inds:
            n_syn[ind] += 1
        return Nsyn, n_syn

    Npre, n_pre = _precompute_synapse_inds(pre_inds)
    Npost, n_post = _precompute_synapse_inds(post_inds)
    
    syn_btwn = np.zeros(len(sk.vertices))
    cov_paths_rev = sk.cover_paths[::-1]
    for path in cov_paths_rev:
        downstream_pre = 0
        downstream_post = 0
        for ind in path:
            downstream_pre += n_pre[ind]
            downstream_post += n_post[ind]
            syn_btwn[ind] = downstream_pre * (Npost - downstream_post) + \
                               (Npre - downstream_pre) * downstream_post
        # Deposit each branch's synapses at the branch point.
        bp_ind = sk.parent_node(path[-1])
        if bp_ind is not None:
            n_pre[bp_ind]  += downstream_pre
            n_post[bp_ind] += downstream_post
    return syn_btwn


def split_axon_by_synapse_betweenness(sk, pre_inds, post_inds, return_quality=True, extend_to_segment=True):
    '''
    Find the is_axon boolean label for all vertices in the skeleton. Assumes skeleton root is not on the axon side.
    :param sk: Skeleton
    :param pre_inds: List of skeleton indices with an input synapse (Duplicate indices with more than one)
    :param post_inds: List of skeleton indices with an output synapse (Duplicate indices with more than one)
    :param return_quality: Compute split quality at the split point. Always computer for true split point, not shifted one.
    :param extend_to_segment: Shift split point to the closest-to-root location on the same segment as the split node.
    :returns: boolean array, True for axon vertices.
    :returns: float, optional split quality index.
    '''
    pre_inds = _check_ind_list(pre_inds)
    post_inds = _check_ind_list(post_inds)

    axon_split = find_axon_split_vertex_by_synapse_betweenness(sk, pre_inds, post_inds, return_quality=return_quality, extend_to_segment=True)
    if return_quality:
        axon_split_ind, split_quality = axon_split
    else:
        axon_split_ind = axon_split
    is_axon = np.full(len(sk.vertices), False)
    is_axon[ sk.downstream_nodes(axon_split_ind) ] = True
    
    if return_quality:
        return is_axon, split_quality
    else:
        return is_axon


def find_axon_split_vertex_by_synapse_betweenness(sk, pre_inds, post_inds, return_quality=True, extend_to_segment=True):
    '''
    Find the skeleton vertex at which to split the axon from the dendrite. Assumes skeleton root is on dendritic side.
    :param sk: Skeleton
    :param pre_inds: List of skeleton indices with an input synapse (Duplicate indices with more than one)
    :param post_inds: List of skeleton indices with an output synapse (Duplicate indices with more than one)
    :param return_quality: Compute split quality at the split point. Always computer for true split point, not shifted one.
    :param extend_to_segment: Shift split point to the closest-to-root location on the same segment as the split node.
    :returns: int, skeleton index
    :returns: float, optional split quality index.
    '''
    pre_inds = _check_ind_list(pre_inds)
    post_inds = _check_ind_list(post_inds)

    syn_btwn = synapse_betweenness(sk, pre_inds, post_inds)
    high_vinds = np.flatnonzero(syn_btwn==max(syn_btwn))
    close_vind = high_vinds[np.argmin(sk.distance_to_root[high_vinds])]
    
    if return_quality:
        axon_qual_label = np.full(len(sk.vertices), False)
        axon_qual_label[ sk.downstream_nodes(close_vind) ] = True
        split_quality = axon_split_quality(axon_qual_label, pre_inds, post_inds)

    if extend_to_segment:
        relseg = sk.segment_map[close_vind]
        axon_split_ind = sk.segments[relseg][-1]
    else:
        axon_split_ind = close_vind
    
    if return_quality:
        return axon_split_ind, split_quality
    else:
        return axon_split_ind


def axon_split_quality(is_axon, pre_inds, post_inds):
    '''
    Returns a quality index (0-1, higher is a cleaner split) for split quality,
    defined as best separating input and output sites.
    '''
    pre_inds = _check_ind_list(pre_inds)
    post_inds = _check_ind_list(post_inds)

    axon_pre = sum(is_axon[pre_inds])
    axon_post = sum(is_axon[post_inds])
    dend_pre = sum(~is_axon[pre_inds])
    dend_post = sum(~is_axon[post_inds])

    counts = np.array([[axon_pre, axon_post],[dend_pre, dend_post]])
    observed_ent = _distribution_split_entropy( counts )
    
    unsplit_ent = _distribution_split_entropy([[len(pre_inds), len(post_inds)]] )

    return 1-observed_ent/unsplit_ent


def _distribution_split_entropy(counts):
    if np.sum(counts)==0:
        return 0
    ps = np.divide(counts, np.sum(counts, axis=1)[:,np.newaxis], where=np.sum(counts, axis=1)[:,np.newaxis]>0)
    Hpart = np.sum(np.multiply(ps, np.log2(ps, where=ps>0)), axis=1)
    Hws = np.sum(counts, axis=1) / np.sum(counts)
    Htot = -np.sum(Hpart * Hws)
    return Htot


def _check_ind_list(inds):
    if type(inds) is dict:
        return np.concatenate([[k]*len(v) for k,v in inds.items() if k>=0])
    else:
        return inds


__doc__ = """
Mesh Iterator Classes
"""

import random
import time
import numpy as np


ORDERS = ["random", "sequential"]


class LocalViewIterator(object):
    """
    Iterator class which samples local views that cover an entire mesh.
    Each mesh vertex is counted as "covered" if it's included in at
    least one local view across the iterator.
    """

    def __init__(self, mesh, n_points, batch_size=1, order="random",
                 pc_align=False, pc_norm=False, adaptnorm=False,
                 fisheye=False, sample_n_points=None, verbose=False):

        assert order in ORDERS, f"invalid order {order} not in {ORDERS}"

        self._active_inds = list(range(mesh.vertices.shape[0]))
        self._order = order
        self._mesh = mesh
        self._batch_size = batch_size

        # arguments for local view method calls
        self._kwargs = dict(n_points=n_points, pc_align=pc_align,
                            fisheye=fisheye,
                            verbose=verbose, sample_n_points=sample_n_points,
                            return_node_ids=True, pc_norm=pc_norm)

        self._deact_kwargs = dict(n_points=n_points, pc_align=pc_align,
                                  adapt_unit_sphere_norm=adaptnorm,
                                  verbose=verbose, sample_n_points=None,
                                  return_node_ids=True, pc_norm=pc_norm)

    def __iter__(self):
        return self

    def __next__(self):
        time_start = time.time()
        # stopping condition: no more indices to sample
        if len(self._active_inds) == 0:
            raise StopIteration

        n_samples = min(self._batch_size, len(self._active_inds))

        if self._order == "random":
            random.seed(time.time())
            centers = np.random.choice(self._active_inds, n_samples,
                                       replace=False)
            views, _, node_ids = self._mesh.get_local_views(
                center_node_ids=centers,  **self._kwargs)

            if self._kwargs["sample_n_points"] is not None:
                _, _, node_ids = self._mesh.get_local_views(
                    center_node_ids=centers, **self._deact_kwargs)

            self._deactivate_nodes(node_ids.flatten())
        elif self._order == "sequential":
            centers = []
            views = []
            while len(self._active_inds) > 0:
                centers.append(self._active_inds[0])
                view, _, node_ids = self._mesh.get_local_views(
                    center_node_ids=[centers[-1]], **self._kwargs)
                views.extend(view)

                if self._kwargs["sample_n_points"] is not None:
                    _, _, node_ids = self._mesh.get_local_views(
                        center_node_ids=centers, **self._deact_kwargs)

                self._deactivate_nodes(node_ids)
        else:
            raise Exception()

        print("Views took %.3fs" % (time.time() - time_start))
        return np.array(views, dtype=np.float32), \
               np.array(centers, dtype=np.uint32)

    def _deactivate_nodes(self, node_ids):
        """
        Removes nodes from consideration which have been sampled
        in the last patch
        """
        if isinstance(node_ids, np.ndarray):
            node_ids = node_ids.tolist()

        to_deactivate = set(node_ids)
        self._active_inds = list(filter(lambda x: x not in to_deactivate,
                                        self._active_inds))



import numpy as np
from scipy import sparse


def filter_close_to_line(mesh, line_end_pts, line_dist_th, axis=1, endcap_buffer=0, sphere_ends=False, map_to_unmasked=True):
    '''
    Given a mesh and a line segment defined by two end points, make a filter
    leaving only those nodes within a certain distance of the line segment in
    a plane defined by a normal axis (e.g. the y axis defines distances in the
    xy plane)

    Parameters
    ----------
    mesh : meshparty.trimesh_io.Mesh
        Trimesh-like mesh with N vertices
    line_end_pts: numpy.array
        2x3 numpy array defining the two end points
    line_dist_th: numeric
        numeric, distance threshold
    axis: int
        integer 0-2. Defines which axis is normal to the plane in
        which distances is computed. optional, default 1 (y-axis).

    Returns
    -------
    numpy.array
        N-length boolean array

    '''
    line_pt_ord = np.argsort(line_end_pts[:, axis])
    ds = _dist_from_line(mesh.vertices, line_end_pts, axis)

    below_top = mesh.vertices[:, axis] > line_end_pts[line_pt_ord[0], axis] - endcap_buffer
    above_bot = mesh.vertices[:, axis] < line_end_pts[line_pt_ord[1], axis] + endcap_buffer
    is_close = (ds < line_dist_th) & above_bot & below_top

    if sphere_ends is True:
        near_a = np.linalg.norm(mesh.vertices - line_end_pts[0], axis=1) < line_dist_th
        near_b = np.linalg.norm(mesh.vertices - line_end_pts[1], axis=1) < line_dist_th
        end_cap = near_a | near_b
        is_close = is_close | end_cap

    if map_to_unmasked:
        is_close = mesh.map_boolean_to_unmasked(is_close)
    return is_close


def _dist_from_line(pts, line_end_pts, axis):
    ps = (pts[:, axis] - line_end_pts[0, axis]) / (line_end_pts[1, axis] - line_end_pts[0, axis])
    line_pts = np.multiply(ps[:, np.newaxis], line_end_pts[1] - line_end_pts[0]) + line_end_pts[0]
    ds = np.linalg.norm(pts - line_pts, axis=1)
    return ds


def filter_components_by_size(mesh, min_size=0, max_size=np.inf, map_to_unmasked=True):
    """
    returns a boolean mask for vertices that are part of components in a size range

    Parameters
    ----------
    mesh : meshparty.trimesh_io.Mesh
        A Trimesh-like mesh with N vertices
    min_size : int
        the minimum number of vertices in compoment (default 0)
    max_size : int
        the maximum number of vertices in compoment (default infinity)

    Returns
    -------
    np.array
        N-length boolean array

    """
    cc, labels = sparse.csgraph.connected_components(mesh.csgraph, directed=False)
    uids, counts = np.unique(labels, return_counts=True)
    good_labels = uids[(counts > min_size) & (counts <= max_size)]
    is_good = np.in1d(labels, good_labels)
    if map_to_unmasked:
        is_good = mesh.map_boolean_to_unmasked(is_good)
    return is_good


def filter_largest_component(mesh, map_to_unmasked=True):
    """ returns a boolean mask for vertices that are part of the largest component

    Parameters
    ----------
    mesh : meshparty.trimesh_io.Mesh
        A Trimesh-like mesh with N vertices

    Returns
    -------
    np.array
        N-length boolean array

    """
    cc, labels = sparse.csgraph.connected_components(mesh.csgraph)
    uids, counts = np.unique(labels, return_counts=True)
    max_label = np.argmax(counts)
    in_largest = labels == max_label
    if map_to_unmasked:
        in_largest = mesh.map_boolean_to_unmasked(in_largest)
    return in_largest


def filter_spatial_distance_from_points(mesh, pts, d_max, map_to_unmasked=True):
    """
    returns a boolean mask for vertices near a set of points

    Parameters
    ----------
    mesh : meshparty.trimesh_io.Mesh
        A Trimesh-like mesh with N vertices
    pts : numpy.array
        a Kx3 set of points
    d_max : float
        the maximum distance to points to include (same units as mesh.vertices)

    Returns
    -------
    np.array
        N-length boolean array

    """
    if type(pts) == list:
        pts = np.array(pts)
    if len(pts.shape) == 1:
        assert(len(pts) == 3)
        ds = np.linalg.norm(mesh.vertices-pts[np.newaxis, :], axis=1)
        return ds < d_max
    close_enough = np.full((len(mesh.vertices), len(pts)), False)
    for ii, pt in enumerate(pts):
        ds = np.linalg.norm(mesh.vertices-pt, axis=1)
        close_enough[:, ii] = ds < d_max

    is_close = np.any(close_enough, axis=1)
    if map_to_unmasked:
        is_close = mesh.map_boolean_to_unmasked(is_close)
    return is_close

def filter_geodesic_distance(mesh, points, max_distance, max_valid_mapping=np.inf, map_to_unmasked=True):
    '''
    Returns a boolean array of mesh points within a max distance of points along the mesh graph.

    Parameters
    ----------
    mesh : meshparty.trimesh_io.Mesh
        A Trimesh-like mesh with N vertices
    points: numpy.array
        An Mx3 array of points in space or M-length array of mesh indices
    max_distance : float
        Max distance along the mesh graph to include in the filter.
    max_valid_mapping : float
        If points are used, sets the max distance for valid mesh point mapping.
    map_to_unmasked : bool
        If True, returns mask in indexing of the Nu-length unmasked mesh.

    Returns
    -------
    mask : np.array
        Boolean array with Nu (or N) entries, True where vertices are close to any of the points/indices provided.
    '''
    points = np.array(points)
    if len(points.shape) == 2:
        if points.shape[1] == 3:
            ds, inds = mesh.kdtree.query(points, distance_upper_bound=max_valid_mapping)
            inds = inds[~np.isinf(ds)]
        else:
            inds = points.ravel()
    else:
        inds = points
    return np.invert(np.isinf(sparse.csgraph.dijkstra(mesh.csgraph, indices=inds, limit=max_distance, min_only=True)))



def filter_two_point_distance(mesh, pts_foci, d_pad, indices=None, power=1, map_to_unmasked=True):
    '''
    Returns a boolean array of mesh points such that the sum of the distance from a
    point to each of the two foci are less than a constant. The constant is set by
    the distance between the two foci plus a user-specified padding. Optionally, use
    other Minkowski-like metrics (i.e. x^n + y^n < d^n where x and y are the distances
    to the foci.)

    Parameters
    ----------
    mesh : meshparty.trimesh_io.Mesh
        A Trimesh-like mesh with N vertices
    pts_foci: numpy.array
        2x3 array with the two foci in 3d space.
    d_pad: float
        Extra padding of the threhold distance beyond the distance between foci.
    indices : iterator
        Instead of pts_foci, one can specify a len(2) list of two indices into the mesh.vertices
        default None. Will override pts_foci.
    power : int
        what power to use in Minkowski-like metrics for distance metric.

    Returns
    -------
    np.array
        N-length boolean array

    '''
    if indices is None:
        _, minds_foci = mesh.kdtree.query(pts_foci)
    else:
        minds_foci = np.array(indices)

    if len(minds_foci) != 2:
        print('One or both mesh points were not found')
        return None

    d_foci_to_all = sparse.csgraph.dijkstra(mesh.csgraph,
                                            indices=minds_foci,
                                            unweighted=False,
                                            )
    dmax = d_foci_to_all[0, minds_foci[1]] + d_pad

    if np.isinf(dmax):
        print('Points are not in the same mesh component')
        return None

    if power != 1:
        is_in_ellipse = np.sum(np.power(d_foci_to_all, power), axis=0) < np.power(dmax, power)
    else:
        is_in_ellipse = np.sum(d_foci_to_all, axis=0) < dmax

    if map_to_unmasked:
        is_in_ellipse = mesh.map_boolean_to_unmasked(is_in_ellipse)
    return is_in_ellipse



import collections
import numpy as np
import h5py
from scipy import spatial, sparse
from sklearn import decomposition
try:
    from pykdtree.kdtree import KDTree
except:
    KDTree = spatial.cKDTree
import os
import networkx as nx
import requests
import time
import re
from collections import defaultdict
import warnings
import logging
from functools import wraps
import cloudvolume
from cloudvolume.datasource.precomputed.mesh.multilod import ShardedMultiLevelPrecomputedMeshSource
from multiwrapper import multiprocessing_utils as mu

import trimesh
from trimesh import caching
try:
    from trimesh import exchange
except ImportError:
    from trimesh import io as exchange

from pymeshfix import _meshfix
from tqdm import trange
import DracoPy
from meshparty import utils, trimesh_repair

try:
    from caveclient import infoservice
    allow_framework_client = True
except ImportError:
    logging.warning(
        "Need to pip install caveclient to use dataset_name parameters")
    allow_framework_client = False


class EmptyMaskException(Exception):
    """Raised when applying a mask that has all zeros"""
    pass


def _get_cv_path_from_info(dataset_name, server_address=None, segmentation_type='graphene'):
    """Get the cloudvolume path from a dataset name. Segmentation type should be
       either `graphene` or `flat`.
    """
    if allow_framework_client is False:
        logging.warning(
            "Need to pip install caveclient to use dataset_name parameters")
        return None

    info = infoservice.InfoServiceClient(
        dataset_name=dataset_name, server_address=server_address)
    if segmentation_type == 'graphene':
        cv_path = info.graphene_source(format_for='cloudvolume')
    elif segmentation_type == 'flat':
        cv_path = info.flat_segmentation_source(format_for='cloudvolume')
    else:
        cv_path = None
    return cv_path


def read_mesh_h5(filename):
    """Reads a mesh's vertices, faces and normals from an hdf5 file
    assert's that this file exists.
    Will load normals, link_edges, and node_mask if they exist.

    Parameters
    ----------
    filename: str
        a path to a h5 file

    Returns
    -------
    :obj:`np.array`
        vertices, a Nx3 x,y,z coordinates (float)
    :obj:`np.array`
        faces, a Mx3 a,b,c index into vertices for triangle faces np.int32
    :obj:`np.array`:
        normals, A Mx3 x,y,z direction for face normals, np.float32
        [] if it doesn't exist
    :obj:`np.array`
        link_edges, a Kx2 a,b list of extra link edges to add to the mesh graph np.int32
        None if this doesn't exist
    :obj:`np.array`
        node_mask, a N length np.bool area of whether to mask this index (None if doesn't exist)
        None if this doesn't exist

    Raises
    ------
        AssertionError
            if the filename is not a file
    """
    assert os.path.isfile(filename)

    with h5py.File(filename, "r") as f:
        if "draco" in f.keys():
            mesh_object = DracoPy.decode_buffer_to_mesh(
                f["draco"][()].tostring())
            vertices = np.array(mesh_object.points).astype(np.float32)
            if len(vertices.shape) == 1:
                N = len(vertices)
                vertices = vertices.reshape((N//3, 3))
            faces = np.array(mesh_object.faces).astype(np.uint32)
        else:
            vertices = f["vertices"][()]
            faces = f["faces"][()]

        if len(faces.shape) == 1:
            faces = faces.reshape(-1, 3)

        if "normals" in f.keys():
            normals = f["normals"][()]
        else:
            normals = None

        if "link_edges" in f.keys():
            link_edges = f["link_edges"][()]
        else:
            link_edges = None

        if "node_mask" in f.keys():
            node_mask = f["node_mask"][()]
        else:
            node_mask = None
    return vertices, faces, normals, link_edges, node_mask


def write_mesh_h5(filename, vertices, faces,
                  normals=None, link_edges=None, node_mask=None,
                  draco=False, overwrite=False):
    """Writes a mesh's vertices, faces (and normals) to an hdf5 file

    Parameters
    ----------
    filename: str
        a path to a h5 file to write a mesh
    vertices : np.array
        a Nx3 x,y,z coordinates (float)
    faces: np.array
        a Mx3 a,b,c index into vertices for triangle faces np.int32  
    normals: np.array
        a Mx3 x,y,z direction for face normals, np.float32
        if it doesn't exist (default None)
    link_edges: np.array
        a Kx2 a,b list of extra link edges to add to the mesh graph np.int32
        None if this doesn't exist (default None)
    node_mask: np.array
        a N length np.bool area of whether to mask this index (None if doesn't exist)
        None if this doesn't exist (default None)
    overwrite: False
        whether to overwrite the file, will return silently if mesh file exists already

    """

    if os.path.isfile(filename):
        if overwrite:
            os.remove(filename)
        else:
            return

    with h5py.File(filename, "w") as f:
        if draco:

            buf = DracoPy.encode_mesh_to_buffer(vertices.flatten('C'),
                                                faces.flatten('C'))
            f.create_dataset("draco", data=np.void(buf))
        else:
            f.create_dataset("vertices", data=vertices, compression="gzip")
            f.create_dataset("faces", data=faces, compression="gzip")

        if normals is not None:
            f.create_dataset("normals", data=normals, compression="gzip")

        if link_edges is not None:
            f.create_dataset("link_edges", data=link_edges, compression="gzip")

        if node_mask is not None:
            f.create_dataset("node_mask", data=node_mask, compression="gzip")


def read_mesh(filename):
    """Reads a mesh from obj or h5 file

    Parameters
    ----------
    filename: str
        a path to a obj or h5 file to read a mesh

    Returns
    -------
    :obj:`numpy.array`
        vertices, a Nx3 x,y,z coordinates (float)
    :obj:`numpy.array`
        faces, a Mx3 a,b,c index into vertices for triangle faces np.int32
    :obj:`numpy.array`:
        normals, A Mx3 x,y,z direction for face normals, np.float32
        None if it doesn't exist
    :obj:`numpy.array`
        link_edges, a Kx2 a,b list of extra link edges to add to the mesh graph np.int32
        None if this doesn't exist or is an obj file
    :obj:`numpy.array`
        node_mask, a N length np.bool area of whether to mask this index (None if doesn't exist)
        None if this doesn't exist or is an obj file

    """

    if filename.endswith(".obj"):
        with open(filename, 'r') as fp:
            mesh_d = exchange.obj.load_obj(fp)
        vertices = mesh_d['vertices']
        faces = mesh_d['faces']
        normals = mesh_d.get('normals', None)
        link_edges = None
        node_mask = None
    elif filename.endswith(".h5"):
        mesh_data = read_mesh_h5(filename)
        vertices, faces, normals, link_edges, node_mask = mesh_data
    else:
        raise Exception("Unknown filetype")
    return vertices, faces, normals, link_edges, node_mask


def _download_meshes_thread_graphene(args):
    """ Helper to Download meshes into target directory from graphene sources.
    Parameters
    ----------
    args : tuple
        seg_ids : iterator of ids
            the seg ids (with filenames = f"{seg_id}.h5") in the target_dir
        cv_path: str
            the cloudvolume path passed to cloudvolume.CloudVolume
        target_dir: str
            a path to the diretory to save the meshes
        fmt: str
            'hdf5', 'obj', 'stl' or any format supported by 'func':`write_to_file`
        overwrite: bool
            whether to overwrite the meshes if they already exist.
            will do no work if those don't exist
        merge_large_components: bool
            whether to merge all the large components using 'func':trimesh_io.Mesh.merge_large_components
            with default parameters (probably should be False)
        stitch_mesh_chunks: bool
            whether to stitch mesh chunks across meshes after downloading fragments (probably should be True)
        map_gs_to_https: bool
            whether to trigger cloudvolume.CloudVolume use_https option. Probably should be true unless you have
            a private bucket and have ~/.cloudvolume/secrets setup properly
        remove_duplicate_vertices: bool
            whether to bluntly merge duplicate vertices (probably should be False)
        chunk_size: tuple
            size of chunks when deduplicating
        save_draco: bool
            whether to save meshes as draco compressed
        progress: bool
            show progress bars

     """
    seg_ids, cv_path, target_dir, fmt, overwrite, \
        merge_large_components, stitch_mesh_chunks, map_gs_to_https, \
        remove_duplicate_vertices, progress, chunk_size, save_draco = args

    cv = cloudvolume.CloudVolume(cv_path, use_https=map_gs_to_https)

    for seg_id in seg_ids:
        print('downloading {}'.format(seg_id))
        target_file = os.path.join(target_dir, f"{seg_id}.h5")
        if not overwrite and os.path.exists(target_file):
            print('file exists {}'.format(target_file))
            continue
        print('file does not exist {}'.format(target_file))

        try:
            cv_mesh = cv.mesh.get(
                seg_id, remove_duplicate_vertices=remove_duplicate_vertices)[seg_id]

            faces = np.array(cv_mesh.faces)
            if len(faces.shape) == 1:
                faces = faces.reshape(-1, 3)

            mesh = Mesh(vertices=cv_mesh.vertices,
                        faces=faces,
                        process=False)

            if merge_large_components:
                mesh.merge_large_components()

            if fmt == "hdf5":
                write_mesh_h5(f"{target_dir}/{seg_id}.h5",
                              mesh.vertices,
                              mesh.faces.flatten(),
                              link_edges=mesh.link_edges,
                              draco=save_draco,
                              overwrite=overwrite)
            else:
                mesh.write_to_file(f"{target_dir}/{seg_id}.{fmt}")
        except Exception as e:
            print(e)


def _download_meshes_thread_precomputed(args):
    """ Helper to Download meshes into target directory

    Parameters
    ----------
    args : tuple
        seg_ids : iterator of ids
            the seg ids (with filenames = f"{seg_id}.h5") in the target_dir
        cv_path: str
            the cloudvolume path passed to cloudvolume.CloudVolume
        target_dir: str
            a path to the diretory to save the meshes
        fmt: str
            'hdf5', 'obj', 'stl' or any format supported by 'func':`write_to_file`
        overwrite: bool
            whether to overwrite the meshes if they already exist.
            will do no work if those don't exist
        merge_large_components: bool
            whether to merge all the large components using 'func':trimesh_io.Mesh.merge_large_components
            with default parameters (probably should be False)
        stitch_mesh_chunks: bool
            whether to stitch mesh chunks across meshes after downloading fragments (probably should be True)
        map_gs_to_https: bool
            whether to trigger cloudvolume.CloudVolume use_https option. Probably should be true unless you have
            a private bucket and have ~/.cloudvolume/secrets setup properly
        remove_duplicate_vertices: bool
            whether to bluntly merge duplicate vertices (probably should be False)
        chunk_size: tuple
            chuck size for deduplification
        progress: bool
            show progress bars
     """
    seg_ids, cv_path, target_dir, fmt, overwrite, \
        merge_large_components, stitch_mesh_chunks, \
        map_gs_to_https, remove_duplicate_vertices, \
        progress, chunk_size, save_draco = args

    cv = cloudvolume.CloudVolume(
        cv_path, use_https=map_gs_to_https,
        progress=progress,
    )

    download_segids = [
        segid for segid in seg_ids
        if overwrite or not os.path.exists(
            os.path.join(target_dir, f"{segid}.h5")
        )
    ]

    already_have = list(set(seg_ids).difference(set(download_segids)))

    if not overwrite:
        print("Already Have: " + str(already_have))

    print("Downloading: " + str(download_segids))

    while len(download_segids):
        download_now = download_segids[:100]
        download_segids = download_segids[len(download_now):]

        cv_meshes = cv.mesh.get(
            download_now,
            remove_duplicate_vertices=remove_duplicate_vertices,
            fuse=False
        )

        for segid, cv_mesh in cv_meshes.items():
            mesh = Mesh(
                vertices=cv_mesh.vertices,
                faces=cv_mesh.faces,
                process=False,
            )

            if merge_large_components:
                mesh.merge_large_components()

            if fmt == "hdf5":
                write_mesh_h5(f"{target_dir}/{segid}.h5",
                              mesh.vertices,
                              mesh.faces.flatten(),
                              link_edges=mesh.link_edges,
                              draco=save_draco,
                              overwrite=overwrite)
            else:
                mesh.write_to_file(f"{target_dir}/{segid}.{fmt}")


def download_meshes(seg_ids, target_dir, cv_path, overwrite=True,
                    n_threads=1, verbose=False,
                    stitch_mesh_chunks=True,
                    merge_large_components=False,
                    remove_duplicate_vertices=False,
                    map_gs_to_https=True, fmt="hdf5",
                    save_draco=False,
                    chunk_size=None,
                    progress=False):
    """ Downloads meshes in target directory (in parallel)
    will break up the seg_ids into n_threads*3 job blocks or fewer and download them all

    Parameters
    ----------
    seg_ids : iterator of ids
            the seg ids (with filenames = f"{seg_id}.h5") in the target_dir
    target_dir: str
        a path to the diretory to save the meshes
    cv_path: str
        the cloudvolume path passed to cloudvolume.CloudVolume
    n_threads: int
        how many parallel processes to use when downloading (default 1)
    overwrite: bool
        whether to overwrite the meshes if they already exist.
        will do no work if those don't exist (default True)
    stitch_mesh_chunks: bool
        whether to stitch mesh chunks across meshes after downloading fragments (default True)
    merge_large_components: bool
        whether to merge all the large components using 'func':trimesh_io.Mesh.merge_large_components
        with default parameters (default False)
    remove_duplicate_vertices: bool
        whether to bluntly merge duplicate vertices (default False)
    map_gs_to_https: bool
        whether to trigger cloudvolume.CloudVolume use_https option. Probably should be true unless you have
        a private bucket and have ~/.cloudvolume/secrets setup properly (default True)
    chunk_size: np.array
        chunk size in nm to use in deduplification (default None)
    fmt: str
        'hdf5', 'obj', 'stl' or any format supported by :func:`meshparty.trimesh_io.Mesh.write_to_file` (default 'hdf5')
    progress: bool
    """

    if n_threads > 1:
        n_jobs = n_threads * 3
    else:
        n_jobs = 1

    if len(seg_ids) < n_jobs:
        n_jobs = len(seg_ids)

    # Use the cv path to establish if the source is graphene or not.
    if re.search('^graphene://', cv_path) is not None:
        _download_meshes_thread = _download_meshes_thread_graphene
    else:
        _download_meshes_thread = _download_meshes_thread_precomputed

    seg_id_blocks = np.array_split(seg_ids, n_jobs)

    multi_args = []
    for seg_id_block in seg_id_blocks:
        multi_args.append([seg_id_block, cv_path, target_dir, fmt,
                           overwrite, merge_large_components, stitch_mesh_chunks,
                           map_gs_to_https, remove_duplicate_vertices, progress, chunk_size, save_draco])

    if n_jobs == 1:
        mu.multiprocess_func(_download_meshes_thread,
                             multi_args, debug=True,
                             verbose=verbose, n_threads=n_threads)
    else:
        mu.multisubprocess_func(_download_meshes_thread,
                                multi_args, n_threads=n_threads,
                                package_name="meshparty", n_retries=40)


class MeshMeta(object):
    """ Manager class to keep meshes in memory and seemingless download them

        Parameters
        ----------
        cache_size: int
            number of meshes to keep in memory adapt this to your available memory and size of meshes
            set to zero to use less memory but read from disk cache
        cv_path: str
            path to pass to cloudvolume.CloudVolume
        dataset_name: str
            Dataset name to use to get cloudvolume path via infoservice
        server_address: str
            Server address for the infoservice. Uses a default value if None.
        segmentation_type: 'graphene' or 'flat'
            Selects which type of segmentation to use. Graphene is for proofreadable segmentations, flat is for static segmentations.
        disk_cache_path: str
            meshes are dumped to this directory => should be equal to target_dir
            in download_meshes (default None will not cache meshes)
        map_gs_to_https: bool
            whether to change gs paths to https paths, via cloudvolume's use_https option
        voxel_scaling: 3x1 numeric
            Allows a post-facto multiplicative scaling of vertex locations. These values are NOT saved, just used for analysis and visualization.
        """

    def __init__(self, cache_size=400, cv_path=None, dataset_name=None, server_address=None, segmentation_type='graphene',
                 disk_cache_path=None, map_gs_to_https=True, voxel_scaling=None):

        self._mesh_cache = {}
        self._cache_size = cache_size
        if cv_path is None and dataset_name is not None:
            cv_path = _get_cv_path_from_info(
                dataset_name=dataset_name, server_address=server_address, segmentation_type=segmentation_type)
        self._cv_path = cv_path
        self._cv = None
        self._map_gs_to_https = map_gs_to_https
        self._disk_cache_path = disk_cache_path
        self._voxel_scaling = voxel_scaling

        if self.disk_cache_path is not None:
            if not os.path.exists(self.disk_cache_path):
                os.makedirs(self.disk_cache_path)

    @property
    def cache_size(self):
        """the size of the cache"""
        return self._cache_size

    @property
    def cv_path(self):
        """str: the path passed to cloudvolume.CloudVolume"""
        return self._cv_path

    @property
    def disk_cache_path(self):
        """str: the path where meshes are saved"""
        return self._disk_cache_path

    @property
    def cv(self):
        """ cloudvoume.CloudVolume : the cloudvolume object"""
        if self._cv is None and self.cv_path is not None:
            self._cv = cloudvolume.CloudVolume(self.cv_path, parallel=10,
                                               use_https=self._map_gs_to_https)

        return self._cv

    @property
    def voxel_scaling(self):
        """np.array : 3 element vector to rescale mesh vertices"""
        return self._voxel_scaling

    def _filename(self, seg_id, lod=None):
        """ a method to define what path this seg_id will or is saved to

        Parameters
        ----------
        seg_id: np.uint64 or int
            the seg_id to get the filename for
        lod: int or None
            the level of detail oof the mesh

        """
        assert self.disk_cache_path is not None
        if lod is not None:
            return "%s/%d_%d.h5" % (self.disk_cache_path, seg_id, lod)
        else:
            return "%s/%d.h5" % (self.disk_cache_path, seg_id)

    def mesh(self, filename=None, seg_id=None, cache_mesh=True,
             merge_large_components=False,
             stitch_mesh_chunks=True,
             overwrite_merge_large_components=False,
             remove_duplicate_vertices=False,
             force_download=False,
             lod=0,
             voxel_scaling='default'):
        """ Loads mesh either from cache, disk or google storage

        Note, if the mesh is in a cache (memory or disk)
        you will get exactly what was in the cache
        irrespective of the other options you specified
        unless force_download is set, except voxel_scaling which is always applied post-facto.

        Parameters
        ----------
        filename: str
            the full path to a file to load (default None)
        seg_id: uint64
            the mesh_id to get (default None, requires cv_path)
        cache_mesh: bool
            if True: mesh is cached in a dictionary. The user is responsible
            for avoiding a memory overflow (default True)
        merge_large_components: bool
            if True: large (>100 vx) mesh connected components are linked
            and the additional edges strored in .link_edges
            this information is cached as well (default False)
        stitch_mesh_chunks: bool
            if True it will stitch the mesh fragments together into a single graph
            (default True)
        overwrite_merge_large_components: bool
            if True: recalculate large components (default False)
        remove_duplicate_vertices: bool
            whether to bluntly removed duplicate vertices (default False)
        lod: int
            what level of detail to download, only relevent for multi-resolution meshes (default =0 )
        force_download: bool
            whether to force the mesh to be redownloaded from cloudvolume
        voxel_scaling: 3 element numeric or None
            Allows a post-facto multiplicative scaling of vertex locations. These values are NOT saved, just used for analysis and visualization.
            By default, pulls from the value in the meshmeta. 

        Returns
        -------
        :obj:`Mesh`
            The mesh object of this seg_id 

        Raises
        ------
        AssertionError
            if filename is not None, and seg_id and cv_path are not both set
            then it doesn't know how to get your mesh
        """
        if not isinstance(self.cv.mesh, ShardedMultiLevelPrecomputedMeshSource):
            lod = None

        if voxel_scaling == 'default':
            voxel_scaling = self.voxel_scaling

        if filename is not None:
            if filename not in self._mesh_cache:
                mesh_data = read_mesh(filename)
                vertices, faces, normals, link_edges, node_mask = mesh_data
                mesh = Mesh(vertices=vertices, faces=faces, normals=normals,
                            link_edges=link_edges, node_mask=node_mask)

                if cache_mesh and len(self._mesh_cache) < self.cache_size:
                    self._mesh_cache[filename] = mesh
            else:
                mesh = self._mesh_cache[filename]

            if self.disk_cache_path is not None and \
                    overwrite_merge_large_components:
                mesh.write_to_file(self._filename(seg_id, lod=lod))
        else:
            if self.disk_cache_path is not None and force_download is False:
                if os.path.exists(self._filename(seg_id, lod=lod)):
                    mesh = self.mesh(filename=self._filename(seg_id, lod=lod),
                                     cache_mesh=cache_mesh,
                                     merge_large_components=merge_large_components,
                                     overwrite_merge_large_components=overwrite_merge_large_components,
                                     voxel_scaling=voxel_scaling)
                    return mesh
            assert (seg_id is not None and self.cv is not None)
            if seg_id not in self._mesh_cache or force_download is True:
                
                if isinstance(self.cv.mesh, ShardedMultiLevelPrecomputedMeshSource):
                    cv_mesh_d = self.cv.mesh.get(seg_id, lod=lod)
                else:
                    cv_mesh_d = self.cv.mesh.get(
                        seg_id,  remove_duplicate_vertices=remove_duplicate_vertices)
                if isinstance(cv_mesh_d, (dict, collections.defaultdict)):
                    cv_mesh = cv_mesh_d[seg_id]
                else:
                    cv_mesh = cv_mesh_d
                faces = np.array(cv_mesh.faces)
                if (len(faces.shape) == 1):
                    faces = faces.reshape(-1, 3)

                mesh = Mesh(vertices=cv_mesh.vertices,
                            faces=faces)
                if isinstance(self.cv.mesh, ShardedMultiLevelPrecomputedMeshSource):
                    mesh=mesh.process()
                if cache_mesh and len(self._mesh_cache) < self.cache_size:
                    self._mesh_cache[seg_id] = mesh

                if self.disk_cache_path is not None:
                    mesh.write_to_file(self._filename(
                        seg_id, lod=lod), overwrite=force_download)
            else:
                mesh = self._mesh_cache[seg_id]

        mesh.voxel_scaling = voxel_scaling

        if (merge_large_components and (len(mesh.link_edges) == 0)) or \
                overwrite_merge_large_components:
            mesh.merge_large_components()
        return mesh


class Mesh(trimesh.Trimesh):
    """An extension of trimesh.Trimesh class to allow more features

    Parameters
    ----------
    *args : a list of :class:`trimesh.Trimesh` arguments
        the first most commonly used...
        vertices : np.array
            a Nx3 array of x,y,z of vertex positions
        faces: np.array
            a Mx3 array of abc indices into vertices that form triangle faces
        normals: np.array
            a Mx3 array of face normals
    node_mask : np.array
        a N long boolean array of which vertices are masked
    unmasked_size: np.array
        how long the original vertex list is (relevant for masked meshes)
    apply_mask: bool
        whether to apply the node_mask to the result
    link_edges: np.array
        a Kx2 array of indices into vertices that represent extra edges you 
        want to store in the mesh graph
    **kwargs:
        all the other keyword args you want to pass to :class:`trimesh.Trimesh`

    """

    def __init__(self, *args, node_mask=None, unmasked_size=None, apply_mask=False, link_edges=None, voxel_scaling=None, **kwargs):
        if 'vertices' in kwargs:
            vertices_all = kwargs.pop('vertices')
        else:
            vertices_all = args[0]

        if 'faces' in kwargs:
            faces_all = kwargs.pop('faces')
        else:
            # If faces are in args, vertices must also have been in args
            faces_all = args[1]

        if unmasked_size is None:
            if node_mask is not None:
                unmasked_size = len(node_mask)
            else:
                unmasked_size = len(vertices_all)
        if unmasked_size < len(vertices_all):
            raise ValueError(
                'Original size cannot be smaller than current size')
        self._unmasked_size = unmasked_size

        if node_mask is None:
            node_mask = np.full(self.unmasked_size, True, dtype=bool)
        elif node_mask.dtype is not np.dtype('bool'):
            node_mask_inds = node_mask.copy()
            node_mask = np.full(self.unmasked_size, False, dtype=bool)
            node_mask[node_mask_inds] = True

        if len(node_mask) != unmasked_size:
            raise ValueError(
                'The node mask must be the same length as the unmasked size')

        self._node_mask = node_mask

        if apply_mask:
            if any(self.node_mask == False):
                nodes_f = vertices_all[self.node_mask]
                faces_f = utils.filter_shapes(
                    np.flatnonzero(node_mask), faces_all)[0]
            else:
                nodes_f, faces_f = vertices_all, faces_all
        else:
            nodes_f, faces_f = vertices_all, faces_all

        new_args = (nodes_f, faces_f)
        if len(args) > 2:
            new_args += args[2:]
        if kwargs.get('process', False):
            print('No silent changing of the mesh is allowed')
        kwargs['process'] = False

        self._voxel_scaling = None
        self._MeshIndex = None

        super(Mesh, self).__init__(*new_args, **kwargs)
        if apply_mask:
            if link_edges is not None:
                if any(self.node_mask == False):
                    self.link_edges = utils.filter_shapes(
                        np.flatnonzero(node_mask), link_edges)[0]
                else:
                    self.link_edges = link_edges
            else:
                self.link_edges = None
        else:
            self.link_edges = link_edges

        self._index_map = None

        self.voxel_scaling = voxel_scaling

    # Helper class for handling scaling issues

    class ScalingManagement(object):
        @staticmethod
        def original_scaling(func):
            @wraps(func)
            def wrapper(self, *args, **kwargs):
                original_scaling = self.voxel_scaling
                self.voxel_scaling = None
                func(self, *args, **kwargs)
                self.voxel_scaling = original_scaling
            return wrapper

    @property
    def voxel_scaling(self):
        return self._voxel_scaling

    @voxel_scaling.setter
    def voxel_scaling(self, new_scaling):
        self._update_voxel_scaling(new_scaling)

    @property
    def inverse_voxel_scaling(self):
        if self.voxel_scaling is not None:
            return 1/self.voxel_scaling
        else:
            return None

    def _update_voxel_scaling(self, new_scaling):
        """Update the scale of the mesh

        Parameters
        ----------
        new_scale : 3-element vector 
            Sets the new xyz scale relative to the resolution from the mesh source
        """
        if self.voxel_scaling is not None:
            self.vertices = self.vertices * self.inverse_voxel_scaling

        if new_scaling is not None:
            self._voxel_scaling = np.array(new_scaling).reshape(3)
            self.vertices = self.vertices * self._voxel_scaling
        else:
            self._voxel_scaling = None

        self._clear_extra_cached_vertex_keys()

    def _clear_extra_cached_vertex_keys(self, keys=['nxgraph', 'csgraph', 'pykdtree', 'kdtree']):
        for k in keys:
            self._cache.delete(k)

    @property
    def link_edges(self):
        """numpy.array : a Kx2 set of extra edges you want to store in the mesh graph,
        :func:`edges` will return this plus :func:`face_edges`"""
        return self._data['link_edges']

    @link_edges.setter
    def link_edges(self, values):
        """this will invalidate the cached properties that are graph related"""
        if values is None:
            values = np.array([[], []]).T
        values = np.asanyarray(values, dtype=np.int64)
        # prevents cache from being invalidated
        with self._cache:
            self._data['link_edges'] = values
        # now invalidate all items affected
        # not sure this is all of them that are not affected
        # by adding link_edges
        self._cache.clear(exclude=['face_normals',
                                   'vertex_normals',
                                   'faces_sparse',
                                   'bounds',
                                   'extents',
                                   'scale',
                                   'centroid',
                                   'principal_inertia_components',
                                   'principal_inertia_transform',
                                   'symmetry',
                                   'triangles',
                                   'triangles_tree',
                                   'triangles_center',
                                   'triangles_cross',
                                   'edges',
                                   'edges_face',
                                   'edges_unique',
                                   'edges_unique_length'])

    @caching.cache_decorator
    def nxgraph(self):
        """:class:`networkx.Graph` : networkx graph of the mesh"""
        return self._create_nxgraph()

    @caching.cache_decorator
    def csgraph(self):
        """:mod:`scipy.sparse.csgraph` : graph of the mesh"""
        return self._create_csgraph()

    @caching.cache_decorator
    def pykdtree(self):
        """pykdtree.KDTree : KDTree of the mesh vertices"""
        return KDTree(self.vertices)

    @caching.cache_decorator
    def kdtree(self, balanced_tree=False):
        """scipy.spatial.cKDTree : kdtree of the mesh vertices

        Parameters
        ----------
        balanced_tree: bool
            passed on to scipy.spatial.cKDTree
        """
        return spatial.cKDTree(self.vertices, balanced_tree=False)

    @property
    def n_vertices(self):
        """int : how many vertices are in the mesh"""
        return len(self.vertices)

    @property
    def n_faces(self):
        """int : how many faces are in the mesh"""
        return len(self.faces)

    @caching.cache_decorator
    def graph_edges(self):
        # mesh.edges has bidirectional edges, so we need to pass bidirectional link_edges.
        if len(self.link_edges) > 0:
            link_edges_sym = np.vstack(
                (self.link_edges, self.link_edges[:, [1, 0]]))
            link_edges_sym_unique = np.unique(link_edges_sym, axis=1)
        else:
            link_edges_sym_unique = self.link_edges
        return np.vstack([self.edges, link_edges_sym_unique])

    def fix_mesh(self, wiggle_vertices=False, verbose=False):
        """ Executes rudimentary fixing function from pymeshfix

        Good for closing holes, fixes mesh in place
        will recalculate normals

        Parameters
        ----------
        wiggle_vertices: bool
            adds robustness for smaller components (default False)
        verbose: bool
            whether to print out debug statements (default False)
        """
        if self.body_count > 1:
            tin = _meshfix.PyTMesh(verbose)
            # tin.LoadArray(self.vertices, self.faces)
            tin.load_array(self.vertices, self.faces)
            tin.remove_smallest_components()
            # tin.RemoveSmallestComponents()

            # Check if volume is 0 after isolated components have been removed
            # self.vertices, self.faces = tin.ReturnArrays()
            self.vertices, self.faces = tin.return_arrays()

            self.fix_normals()

        if self.volume == 0:
            return

        if wiggle_vertices:
            wiggle = np.random.randn(self.n_vertices * 3).reshape(-1, 3) * 10
            self.vertices += wiggle

        # self.vertices, self.faces = _meshfix.CleanFromVF(self.vertices,
        #                                                  self.faces,
        #                                                  verbose=verbose)
        self.vertices, self.faces = _meshfix.clean_from_arrays(
            self.vertices, self.faces, verbose=verbose)

        self.fix_normals()

    def get_local_views(self, n_points=None,
                        max_dist=np.inf,
                        sample_n_points=None,
                        fisheye=False,
                        pc_align=False,
                        center_node_ids=None,
                        center_coords=None,
                        verbose=False,
                        return_node_ids=False,
                        svd_solver="auto",
                        return_faces=False,
                        adapt_unit_sphere_norm=False,
                        pc_norm=False):
        """ Extracts a local view (points)

        Parameters
        ----------
        n_points: int
            number of points to sample, default None will use the full mesh vertices
        max_dist: float
            sets an upper limit for distance of any sampled mesh point. Might
            reduce n_points
        sample_n_points: int
            has to be >= n_points; if > n_points more points are sampled and a
            subset randomly chosen
        fisheye: bool
            addition to sample_n_points; subset is sampled such that a fisheye
            effect is generated (default False)
        pc_align: bool
            computes PCA and orients mesh along PCs (default False)
        center_node_ids: list of ints
            mesh vertices at the center of the local views (default None)
        center_coords: list (n, 3) of floats
            coordinates at the center of the local views (default None)
            will override center_node_ids
            if both center_node_ids and center_coors are None will choose random spots
            equal to the length of vertices
        verbose: bool
            whether to print more debugging (default False)
        return_node_ids: bool
            sampled node ids are returned as well, changes the output format (default False)
        svd_solver: str
            PCA solver passed to sklearn.decomposition.PCA (default "auto")
        return_faces: bool
            sampled faces are returned as well, changes the output format (default False)
        adapt_unit_sphere_norm: bool
            NOT FUNCTIONAL (default False)
        pc_norm: bool
            if True: normalize point cloud to mean 0 and std 1 before PCA (default False)

        Returns
        -------
        np.array
            local_vertices, a list of n_points or n_sample_points x 3 matrix of points
            len(local_vertices)=K will depend on center_coords or center_coords_ind
        np.array
            center_node_ids, a K long array of center node ids.. useful if you had it choose random points
            if you had passed center_coords this will not accurately reflect the centers used
        np.array
            return_node_ids, Optional depending on whether return_node_ids. A K long list of 
        np.array
            return_faces, Optional depending on return_faces. faces on the local views, a K list of mx3 triangle faces. 

        """
        if center_node_ids is None and center_coords is None:
            center_node_ids = np.array([np.random.randint(len(self.vertices))])

        if center_coords is None:
            center_node_ids = np.array(center_node_ids, dtype=np.int)
            center_coords = self.vertices[center_node_ids]

        if sample_n_points is None:
            sample_n_points = n_points
        elif sample_n_points > n_points:
            assert not return_faces
        elif sample_n_points == n_points:
            pass
        else:
            raise Exception("Too few sample points specified")

        center_coords = np.array(center_coords)

        if sample_n_points is None:
            sample_n_points = len(self.vertices)
        else:
            sample_n_points = np.min([sample_n_points, len(self.vertices)])

        dists, node_ids = self.kdtree.query(center_coords, sample_n_points,
                                            distance_upper_bound=max_dist)

        if n_points is not None:
            if sample_n_points > n_points:
                if fisheye:
                    probs = 1 / dists

                    new_dists = []
                    new_node_ids = []
                    ids = np.arange(0, sample_n_points, dtype=np.int)
                    for i_sample in range(len(center_coords)):
                        sample_ids = np.random.choice(ids, n_points,
                                                      replace=False,
                                                      p=probs[i_sample])
                        new_dists.append(dists[i_sample, sample_ids])
                        new_node_ids.append(node_ids[i_sample, sample_ids])

                    dists = np.array(new_dists, dtype=np.float32)
                    node_ids = np.array(new_node_ids, dtype=np.int)
                else:
                    ids = np.arange(0, sample_n_points, dtype=np.int)
                    sample_ids = np.random.choice(ids, n_points, replace=False)

                    dists = dists[:, sample_ids]
                    node_ids = node_ids[:, sample_ids]

        if verbose:
            print(np.mean(dists, axis=1), np.max(dists, axis=1),
                  np.min(dists, axis=1))

        if max_dist < np.inf:
            node_ids = list(node_ids)
            local_vertices = []
            for i_ns, ns in enumerate(node_ids):
                ns = ns[dists[i_ns] <= max_dist]
                ns = np.sort(ns)
                node_ids[i_ns] = ns
                local_vertices.append(self.vertices[ns].copy())
        else:
            node_ids = np.sort(node_ids, axis=1)
            local_vertices = self.vertices[node_ids].copy()

        if pc_align:
            for i_lv in range(len(local_vertices)):
                local_vertices[i_lv] = self._calc_pc_align(local_vertices[i_lv],
                                                           svd_solver,
                                                           pc_norm=pc_norm)

        if adapt_unit_sphere_norm:
            local_vertices -= center_coords
            lengths = np.linalg.norm(local_vertices, axis=2)
            local_vertices /= np.max(lengths, axis=1)[:, None, None]

        return_tuple = (local_vertices, center_node_ids)

        if return_node_ids:
            return_tuple += (node_ids, )

        if return_faces:
            return_tuple += (self._filter_faces(node_ids), )

        return return_tuple

    def get_local_view(self, n_points=None, max_dist=np.inf,
                       sample_n_points=None,
                       adapt_unit_sphere_norm=False,
                       fisheye=False,
                       pc_align=False, center_node_id=None,
                       center_coord=None, method="kdtree", verbose=False,
                       return_node_ids=False, svd_solver="auto",
                       return_faces=False, pc_norm=False):
        """ Single version of get_local_views for backwards compatibility """

        assert method == "kdtree"

        if center_node_id is None and center_coord is None:
            center_node_id = np.random.randint(len(self.vertices))

        if center_coord is None:
            center_coord = self.vertices[center_node_id]

        return self.get_local_views(n_points=n_points,
                                    sample_n_points=sample_n_points,
                                    max_dist=max_dist,
                                    pc_align=pc_align,
                                    center_node_ids=[center_node_id],
                                    center_coords=[center_coord],
                                    adapt_unit_sphere_norm=adapt_unit_sphere_norm,
                                    fisheye=fisheye,
                                    verbose=verbose,
                                    return_node_ids=return_node_ids,
                                    svd_solver=svd_solver,
                                    return_faces=return_faces,
                                    pc_norm=pc_norm)

    def _filter_faces(self, node_ids):
        """ method to return reindexed faces that involve only certain vertices

        Parameters
        ----------
        node_ids: np.array
            a M long set of indices into vertices that you want to filter faces by
            so only return faces that involve these vertices. node_ids has to be sorted! 

        Returns
        -------
        np.array 
            a Kx3 matrix that is a proper faces for a mesh whose vertices = mesh.vertices[node_ids]
        """
        return utils.filter_shapes(node_ids, self.faces)

    def _filter_graph_edges(self, node_ids):
        """ method to return reindexed edges that involve only certain vertices

        Parameters
        ----------
        node_ids: np.array
            a M long set of indices into vertices that you want to filter graph_edges by
            so only return faces that involve these vertices. node_ids has to be sorted! 

        Returns
        -------
        np.array 
            a Kx2 matrix of edges for a mesh whose vertices = mesh.vertices[node_ids]
        """

        return utils.filter_shapes(node_ids, self.graph_edges)

    @ScalingManagement.original_scaling
    def add_link_edges(self, seg_id=None, merge_log=None, datastack_name=None, server_address=None,
                       close_map_distance=300, client=None, verbose=False, base_resolution=None):
        """ add a set of link edges to this mesh from a PyChunkedGraph endpoint
        This will ask the pcg server where merges were done and try to calculate 
        where edges should be added to reflect the merge operations that have been done
        on this mesh, linking disconnected portions of the mesh.

        Parameters
        ----------
        seg_id: int 
            the seg_id of this mesh
        merge_log : dict
            JSON dict of merge log as it comes out of the chunkedgraph client. If used, must
            also set base_resolution, the mip 0 resolution of the supervoxel segmentation volume.
        dataset_name: str or None, optional
            The datastack name this mesh can be found in. If None, requires a pre-made client
            passed through the client parameter. Defaults to None.
        close_map_distance: float, optional
            The distance in nm in mesh vertex coordinates to consider a mapping to be 'close'.
            Defaults to 300.
        server_address: str or None, optional
            the server address to find the pcg endpoint (defaults to None)
        client : annotationframeworkclient.FrameworkClient or None, optional
            Framework client for a specific datastack. If provided, ingores datastack name and
            server_address parameters. defaults to None
        verbose : bool, optional
            If True, provides more debugging statements, default is False
        base_resolution : array-like or None, optional
            Resolution of the supervoxel segmentation at its lowest mip.
        """
        if seg_id is None and merge_log is None:
            raise ValueError(
                'Must set either seg id or pre-determined merge log')

        if merge_log is not None:
            link_edges = trimesh_repair.merge_log_edges(self,
                                                        merge_log=merge_log,
                                                        base_resolution=base_resolution,
                                                        close_map_distance=close_map_distance,
                                                        verbose=verbose)
        else:
            # Use the get_link_edges approach
            link_edges = trimesh_repair.get_link_edges(self, seg_id, datastack_name=datastack_name,
                                                       close_map_distance=close_map_distance,
                                                       server_address=server_address,
                                                       verbose=verbose,
                                                       client=client)

        self.link_edges = np.vstack([self.link_edges, link_edges])

    def get_local_meshes(self, n_points, max_dist=np.inf, center_node_ids=None,
                         center_coords=None, pc_align=False, pc_norm=False,
                         fix_meshes=False):
        """ Extracts a local mesh

        Parameters
        ----------
        n_points: int
        max_dist: float
        enter_node_ids: list of ints
        center_coords: list (n, 3) of floats
        pc_align: bool
        pc_norm: bool
        fix_meshes: bool
        """
        local_view_tuple = self.get_local_views(n_points=n_points,
                                                max_dist=max_dist,
                                                center_node_ids=center_node_ids,
                                                center_coords=center_coords,
                                                return_faces=True,
                                                pc_align=pc_align,
                                                pc_norm=pc_norm)
        vertices, _, faces = local_view_tuple

        meshes = [Mesh(vertices=v, faces=f) for v, f in
                  zip(vertices, faces)]

        if fix_meshes:
            for mesh in meshes:
                if mesh.n_vertices > 0:
                    mesh.fix_mesh(wiggle_vertices=False)

        return meshes

    def get_local_mesh(self, n_points=None, max_dist=np.inf,
                       center_node_id=None,
                       center_coord=None, pc_align=True, pc_norm=False):
        """ Single version of get_local_meshes """

        if center_node_id is not None:
            center_node_id = [center_node_id]

        if center_coord is not None:
            center_coord = [center_coord]

        return self.get_local_meshes(n_points, max_dist=max_dist,
                                     center_node_ids=center_node_id,
                                     center_coords=center_coord,
                                     pc_align=pc_align, pc_norm=pc_norm)[0]

    def _calc_pc_align(self, vertices, svd_solver, pc_norm=False):
        """ Calculates PC alignment """

        vertices = vertices.copy()
        if pc_norm:
            vertices -= vertices.mean(axis=0)
            vertices /= vertices.std(axis=0)

        pca = decomposition.PCA(n_components=3,
                                svd_solver=svd_solver,
                                copy=False)

        return pca.fit_transform(vertices)

    def merge_large_components(self, size_threshold=100, max_dist=1000,
                               dist_step=100):
        """ Finds edges between disconnected components
        will add the edges to the existing set of link_edges
        or start a set of link_edges if there are None
        Note: can cause self-contacts to be innapropriately merged

        Parameters
        ----------
        size_threshold: int
            will merge components that have more than this many vertices
            (default 100)
        max_dist: float
            will only merge components that are closer than this
            (default 1000 in units of mesh.vertices, usually nm)
        dist_step: int
            will merge by marching in steps to look for things to merge
            this is the distance of each step (default 100 in units of mesh.vertices)

        """
        time_start = time.time()

        ccs = sparse.csgraph.connected_components(self.csgraph)
        ccs_u, cc_sizes = np.unique(ccs[1], return_counts=True)
        large_cc_ids = ccs_u[cc_sizes > size_threshold]

        print(len(large_cc_ids))

        kdtrees = []
        vertex_ids = []
        for cc_id in large_cc_ids:
            m = ccs[1] == cc_id
            vertex_ids.append(np.where(m)[0])
            v = self.vertices[m]
            kdtrees.append(spatial.cKDTree(v))

        add_edges = []
        for i_tree in range(len(large_cc_ids) - 1):
            for j_tree in range(i_tree + 1, len(large_cc_ids)):
                print("%d - %d      " % (i_tree, j_tree), end="\r")

                if np.any(kdtrees[i_tree].query_ball_tree(kdtrees[j_tree], max_dist)):
                    for this_dist in range(dist_step, max_dist + dist_step, dist_step):

                        pairs = kdtrees[i_tree].query_ball_tree(kdtrees[j_tree],
                                                                this_dist)

                        if np.any(pairs):
                            for i_p, p in enumerate(pairs):
                                if len(p) > 0:
                                    add_edges.extend([[vertex_ids[i_tree][i_p],
                                                       vertex_ids[j_tree][v]]
                                                      for v in p])
                            break

        print(f"Adding {len(add_edges)} new edges.")

        self.link_edges = np.vstack([self.link_edges, add_edges])

        print("TIME MERGING: %.3fs" % (time.time() - time_start))

    def _create_nxgraph(self):
        """ Computes networkx graph for this mesh

        Returns
        -------
        :class:`networkx.Graph`
        """
        return utils.create_nxgraph(self.vertices, self.graph_edges, euclidean_weight=True,
                                    directed=False)

    def _create_csgraph(self):
        """ Computes scipy.sparse.csgraph with weights equal to euclidean distance
        with directed=False"""
        return utils.create_csgraph(self.vertices, self.graph_edges, euclidean_weight=True,
                                    directed=True)

    @property
    def node_mask(self):
        '''
        np.array: Returns the node/vertex mask currently applied to the data
        '''
        return self._node_mask

    @property
    def indices_unmasked(self):
        '''
        np.array: Gets the indices of nodes in the filtered mesh in the unmasked index array
        '''
        return np.flatnonzero(self.node_mask)

    @property
    def unmasked_size(self):
        '''
        Returns the unmasked number of nodes in the mesh
        '''
        return self._unmasked_size

    def apply_mask(self, new_mask, **kwargs):
        '''
        Makes a new Mesh by adding a new mask to the existing one.
        new_mask is a boolean array, either of the original vertex space or the
        current masked length (in which case it is padded with zeros appropriately).

        Parameters
        ----------
        new_mask: np.array
            a N long array of bool where False correponds to vertices that should be masked
            N needs to equal to mesh.vertices.shape[0] (or the original vertex shape if you are
            operating on an already masked mesh)
        kwargs: 
            keyword arguments to pass on to the new Mesh.__init__ function

        Returns
        -------
        trimesh_io.Mesh
            the mesh with the mask applied
        '''
        if not np.any(new_mask):
            raise(EmptyMaskException("new_mask is all False, mesh will be empty"))

        # We need to express the mask in the current vertex indices
        if np.size(new_mask) == np.size(self.node_mask):
            joint_mask = self.node_mask & new_mask
            new_mask = self.filter_unmasked_boolean(new_mask)
        elif np.size(new_mask) == self.vertices.shape[0]:
            joint_mask = self.node_mask & self.map_boolean_to_unmasked(
                new_mask)
        else:
            raise ValueError(
                'Incompatible shape. Must be either original length or current length of vertices.')

        if self.voxel_scaling is None:
            new_vertices = self.vertices
        else:
            new_vertices = self.vertices * self.inverse_voxel_scaling
        new_mesh = Mesh(new_vertices,
                        self.faces,
                        node_mask=joint_mask,
                        unmasked_size=self.unmasked_size,
                        voxel_scaling=self.voxel_scaling,
                        **kwargs)
        link_edge_unmask = self.map_indices_to_unmasked(self.link_edges)
        new_mesh._apply_new_mask_in_place(new_mask, link_edge_unmask)
        return new_mesh

    def _apply_new_mask_in_place(self, mask, link_edge_unmask):
        """ Internal function for applying masks.. use apply_mask
        Use builtin Trimesh tools for masking
        The new 0 index is the first nonzero element of the mask.
        Unfortunately, update_vertices maps all masked face values to 0 as well.
        """
        num_zero_expected = np.sum(
            self.faces == np.flatnonzero(mask)[0], axis=1)
        self.update_vertices(mask)

        num_zero_new = np.sum(self.faces == 0, axis=1)
        faces_to_keep = num_zero_new == num_zero_expected
        self.update_faces(faces_to_keep)
        self.link_edges = self.filter_unmasked_indices(link_edge_unmask)

    def map_indices_to_unmasked(self, unmapped_indices):
        '''
        For a set of masked indices, returns the corresponding unmasked indices

        Parameters
        ----------
        unmapped_indices: np.array
            a set of indices in the masked index space

        Returns
        -------
        np.array
            the indices mapped back to the original mesh index space
        '''
        return utils.map_indices_to_unmasked(self.indices_unmasked, unmapped_indices)

    def map_boolean_to_unmasked(self, unmapped_boolean):
        '''
        For a boolean index in the masked indices, returns the corresponding unmasked boolean index

        Parameters
        ----------
        unmapped_boolean : np.array
            a bool array in the masked index space

        Returns
        -------
        np.array
            a bool array in the original index space.  Is True if the unmapped_boolean suggests it should be.
        '''
        return utils.map_boolean_to_unmasked(self.unmasked_size, self.node_mask, unmapped_boolean)

    def filter_unmasked_boolean(self, unmasked_boolean):
        '''
        For an unmasked boolean slice, returns a boolean slice filtered to the masked mesh

        Parameters
        ----------
        unmasked_boolean : np.array
            a bool array in the original mesh index space

        Returns
        -------
        np.array
            returns the elements of unmasked_boolean that are still relevant in the masked index space
        '''
        return utils.filter_unmasked_boolean(self.node_mask, unmasked_boolean)

    def filter_unmasked_indices(self, unmasked_shape, mask=None):
        """
        filters a set of indices in the original mesh space
        and returns it in the masked space

        Parameters
        ----------
        unmasked_shape: np.array
            a set of indices into vertices in the unmasked index space
        mask: np.array or None
            the mask to apply. default None will use this Mesh node_mask

        Returns
        -------
        np.array
            the unmasked_shape indices mapped into the masked index space
        """
        if mask is None:
            mask = self.node_mask
        return utils.filter_unmasked_indices(mask, unmasked_shape)

    def filter_unmasked_indices_padded(self, unmasked_shape, mask=None):
        """
        filters a set of indices in the original mesh space
        and returns it in the masked space

        Parameters
        ----------
        unmasked_shape: np.array
            a set of indices into vertices in the unmasked index space
        mask: np.array or None
            the mask to apply. default None will use this Mesh node_mask

        Returns
        -------
        np.array
            the unmasked_shape indices mapped into the masked index space,
            with -1 where the original index did not map into the masked mesh.
        """
        if mask is None:
            mask = self.node_mask
        return utils.filter_unmasked_indices_padded(mask, unmasked_shape)

    @ScalingManagement.original_scaling
    def write_to_file(self, filename, overwrite=True, draco=False):
        """ Exports the mesh to any format supported by trimesh

        Parameters
        ----------
        filename: str
            the path to where to write the filename. Will use extension to infer format
            '.h5' for hdf5
            '.obj' for wavefront
            all others supported by :func:`trimesh.exchange.export.export_mesh`
        """
        if os.path.splitext(filename)[1] == '.h5':
            write_mesh_h5(filename,
                          self.vertices,
                          self.faces,
                          normals=self.face_normals,
                          link_edges=self.link_edges,
                          node_mask=self.node_mask,
                          draco=draco,
                          overwrite=overwrite)
        else:
            exchange.export.export_mesh(self, filename)

    @property
    def index_map(self):
        '''
        A dict mapping global indices into the masked mesh indices.
        '''
        if self._index_map is None:
            self._index_map = defaultdict(lambda: np.nan)
            for ii, index in enumerate(self.indices_unmasked):
                self._index_map[index] = ii
        return self._index_map


class MaskedMesh(Mesh):
    def __init__(self, *args,  **kwargs):
        warnings.warn(
            "use of MaskedMesh deprecated, Mesh now contains all MaskedMesh functionality",
            DeprecationWarning)
        super(MaskedMesh, self).__init__(*args, **kwargs)



import os
import subprocess
import glob
import numpy as np

from multiwrapper import multiprocessing_utils as mu

HOME = os.path.expanduser("~")

# path_to_scripts = os.path.dirname(__file__) + "/meshlabserver_scripts/"
path_to_scripts = HOME + "/PyChunkedGraph/pychunkedgraph/meshing/meshlabserver_scripts/"


def run_meshlab_script(script_name, arg_dict):
    """ Runs meshlabserver script --headless

    No X-Server required

    :param script_name: str
    :param arg_dict: dict [str: str]
    """
    arg_string = "".join(["-{0} {1} ".format(k, arg_dict[k])
                          for k in arg_dict.keys()])
    command = "xvfb-run --auto-servernum --server-num=1 meshlabserver -s {0}/{1} {2}".\
        format(path_to_scripts, script_name, arg_string)
    p = subprocess.Popen(command, shell=True, stderr=subprocess.PIPE)
    p.wait()


def _run_meshlab_script_on_dir_thread(args):
    script_name, path_block, out_dir, suffix, arg_dict = args

    for path in path_block:
        out_path = "{}/{}{}.obj".format(out_dir,
            "".join(os.path.basename(path).split(".")[:-1]), suffix)

        this_arg_dict = {"i": path, "o": out_path}
        this_arg_dict.update(arg_dict)

        run_meshlab_script(script_name, this_arg_dict)


def run_meshlab_script_on_dir(script_name, in_dir, out_dir, suffix, arg_dict={},
                      n_threads=1):
    paths = glob.glob(in_dir + "/*.obj")

    print(len(paths))

    if len(suffix) > 0:
        suffix = "_{}".format(suffix)

    n_jobs = n_threads * 3
    if len(paths) < n_jobs:
        n_jobs = len(paths)

    path_blocks = np.array_split(paths, n_jobs)

    multi_args = []
    for path_block in path_blocks:
        multi_args.append([script_name, path_block, out_dir, suffix, arg_dict])

    if n_threads == 1:
        mu.multiprocess_func(_run_meshlab_script_on_dir_thread,
                             multi_args, debug=True,
                             verbose=True, n_threads=1)
    else:
        mu.multisubprocess_func(_run_meshlab_script_on_dir_thread,
                                multi_args, n_threads=n_threads)




import blosc
import numpy as np
import numba
from scipy import sparse
from ..trimesh_io import Mesh
from ..skeleton import Skeleton

DEFAULT_VOXEL_RESOLUTION = [4, 4, 40]


def MeshworkIndexFactory(mw):
    class JointMeshIndex(np.ndarray):
        def __new__(cls, mesh_indices):
            if (
                np.array(mesh_indices).dtype is np.dtype("bool")
                and len(mesh_indices) == mw.mesh.n_vertices
            ):
                mesh_indices = np.flatnonzero(mesh_indices)

            mesh_inds = np.asarray(mesh_indices, dtype=int)
            obj = mesh_inds.view(cls)
            obj._mesh_indices_base = mw.mesh.map_indices_to_unmasked(mesh_inds)
            return obj

        def __array_finalize__(self, obj):
            if obj is None:
                return
            self._mesh_indices_base = getattr(obj, "_mesh_indices_base", np.array([]))

        def __getitem__(self, k):
            ret = super(JointMeshIndex, self).__getitem__(k)
            if not isinstance(ret, np.integer):
                ret._mesh_indices_base = self._mesh_indices_base[k]
            return ret

        def __eq__(self, other):
            return np.array(self) == other

        def __lt__(self, other):
            return np.array(self) < other

        def __le__(self, other):
            return np.array(self) <= other

        def __gt__(self, other):
            return np.array(self) > other

        def __ge__(self, other):
            return np.array(self) >= other

        @property
        def to_array(self):
            return np.array(self)

        @property
        def to_mesh_index(self):
            return self

        @property
        def to_mesh_index_base(self):
            return self._mesh_indices_base

        @property
        def to_mesh_index_point(self):
            return self.to_skel_index.to_mesh_region_point

        @property
        def to_mesh_mask_base(self):
            mask = np.full(len(mw.mesh.node_mask), False)
            mask[self.to_mesh_index_base] = True
            return mask

        @property
        def to_mesh_mask(self):
            return mw.mesh.filter_unmasked_boolean(self.to_mesh_mask_base)

        @property
        def to_all_equivalent_mesh(self):
            return self.to_skel_index.to_mesh_index

        @property
        def to_all_equivalent_mask(self):
            return self.to_skel_index.to_mesh_mask

        @property
        def to_skel_index(self):
            if mw.skeleton is None:
                return None
            return JointSkeletonIndex(np.unique(mw._mind_to_skind(self)))

        @property
        def to_skel_index_padded(self):
            if mw.skeleton is None:
                return None
            return mw._mind_to_skind_padded(self)

        @property
        def to_skel_mask(self):
            if mw.skeleton is None:
                return None
            return JointSkeletonIndex(np.unique(mw._mind_to_skind(self))).to_skel_mask

    if mw.skeleton is None:
        return JointMeshIndex, np.array

    class JointSkeletonIndex(np.ndarray):
        def __new__(cls, skel_indices):
            if (
                np.array(skel_indices).dtype is np.dtype("bool")
                and len(skel_indices) == mw.skeleton.n_vertices
            ):
                skel_indices = np.flatnonzero(skel_indices)

            skel_inds = np.asarray(skel_indices, dtype=int)
            obj = skel_inds.view(cls)
            obj._skel_indices_base = mw.skeleton.map_indices_to_unmasked(
                skel_inds[skel_inds >= 0]
            )
            return obj

        def __array_finalize__(self, obj):
            if obj is None:
                return
            self._skel_indices_base = getattr(obj, "_skel_indices_base", np.array([]))

        def __getitem__(self, k):
            ret = super(JointSkeletonIndex, self).__getitem__(k)
            if not isinstance(ret, np.integer):
                ret._skel_indices_base = self._skel_indices_base[k]
            return ret

        def __eq__(self, other):
            return np.array(self) == other

        def __lt__(self, other):
            return np.array(self) < other

        def __le__(self, other):
            return np.array(self) <= other

        def __gt__(self, other):
            return np.array(self) > other

        def __ge__(self, other):
            return np.array(self) >= other

        @property
        def to_mesh_index(self):
            return JointMeshIndex(mw._skind_to_mind_index(self[self >= 0]))

        @property
        def to_array(self):
            return np.array(self)

        @property
        def to_mesh_index_base(self):
            return mw.mesh.map_indices_to_unmasked(self.to_mesh_index)

        @property
        def to_mesh_mask_base(self):
            return mw._skind_to_mind_mask_base(self[self >= 0])

        @property
        def to_mesh_mask(self):
            return mw._skind_to_mind_mask(self[self >= 0])

        @property
        def to_mesh_region(self):
            return [
                JointMeshIndex(x) if self[ii] >= 0 else JointMeshIndex([])
                for ii, x in enumerate(mw._skind_regions(self))
            ]

        @property
        def to_mesh_region_point(self):
            out = mw._skind_region_first(self)
            out[self < 0] = -1
            return JointMeshIndex(out)

        @property
        def to_skel_index(self):
            return self[self >= 0]

        @property
        def to_skel_index_padded(self):
            return self

        @property
        def to_skel_index_base(self):
            return self._skel_indices_base

        @property
        def to_skel_mask_base(self):
            mask = np.full(len(mw.skeleton.node_mask), False)
            mask[self._skel_indices_base] = True
            return mask

        @property
        def to_skel_mask(self):
            return mw.skeleton.filter_unmasked_boolean(self.to_skel_mask_base)

    return JointMeshIndex, JointSkeletonIndex


@numba.njit(parallel=True)
def _in1d_items(elements, mask, test_vals):
    maskinds = np.flatnonzero(mask)
    out = np.zeros((mask.sum(), 2), dtype=np.int64)
    for ii in numba.prange(len(maskinds)):
        val_ind = np.flatnonzero(elements[maskinds[ii]] == test_vals)[0]
        out[ii, 0] = np.int64(val_ind)
        out[ii, 1] = np.int64(maskinds[ii])
    return out


def in1d_items(elements, test_vals):
    """For each item in test_vals, finds all indices in elements that match it"""
    out = _in1d_items(elements, np.isin(elements, test_vals), test_vals)
    return [out[out[:, 0] == ii, 1] for ii in range(len(test_vals))]


def in1d_first_item(elements, test_vals):
    """Given a long list of indices, finds one index for each of the indices
    in test_vals (or a -1)"""
    el_mask = np.isin(elements, test_vals)
    vals, ind_mask = np.unique(elements[el_mask], return_index=True)
    out = np.full(len(test_vals), -1)
    _, slots = np.unique(test_vals[np.isin(test_vals, vals)], return_inverse=True)
    out[slots] = np.flatnonzero(el_mask)[ind_mask]
    return out


def _window_matrix(
    branch_points,
    bp_downstream,
    distance_to_root,
    end_paths_to_root,
    width,
    dist_func=None,
):
    """Generate matrix such that Aij is 1 if j is within width of i."""
    has_bp = np.full(len(distance_to_root), -1)
    has_bp[branch_points] = branch_points
    seen = np.full(len(distance_to_root), False)

    row_ind = []
    col_ind = []
    dist = []

    for path in end_paths_to_root:
        path_d = distance_to_root[path]
        for jj in path:
            dfrom = path_d - distance_to_root[jj]
            window = np.abs(dfrom) <= width
            bp_in_window = has_bp[path[window]]
            bp_in_window = bp_in_window[bp_in_window >= 0]
            ind_list = [path[window]]
            dist_list = [dfrom[window]]
            for kk in bp_in_window:
                init_dist = np.abs(distance_to_root[jj] - distance_to_root[kk])
                remaining_dist = width - init_dist
                inds_downstream = bp_downstream[kk]
                downstream_dist = (
                    distance_to_root[inds_downstream] - distance_to_root[kk]
                )
                downstream_window = downstream_dist <= remaining_dist
                inds_to_add = inds_downstream[downstream_window]
                dist_to_add = init_dist + downstream_dist[downstream_window]
                ind_list.append(inds_to_add)
                dist_list.append(dist_to_add)
            all_inds, indices = np.unique(np.concatenate(ind_list), return_index=True)
            all_dist = np.concatenate(dist_list)[indices]
            for kk, d in zip(all_inds, all_dist):
                row_ind.append(jj)
                col_ind.append(kk)
                dist.append(d)
            if seen[jj]:
                break
            else:
                seen[jj] = True
        if dist_func is not None:
            dist = dist_func(np.array(dist))

    return dist, (np.array(row_ind), np.array(col_ind))


def window_matrix(sk, width, dist_func=None):
    """Generate matrix such that Aij is 1 if j is within width of i."""
    bp_downstream = {bp: sk.downstream_nodes(bp) for bp in sk.branch_points}
    end_paths_to_root = [sk.path_to_root(ep) for ep in sk.end_points]
    dist, wm_inds = _window_matrix(
        sk.branch_points, bp_downstream, sk.distance_to_root, end_paths_to_root, width
    )
    if dist_func is not None:
        return sparse.csr_matrix(
            (dist, wm_inds), shape=((sk.n_vertices, sk.n_vertices))
        )
    else:
        return sparse.csr_matrix(
            (np.ones(len(wm_inds[0])), wm_inds), shape=((sk.n_vertices, sk.n_vertices))
        )


def unique_column_name(base_name, suffix, df):
    if base_name is not None:
        col_name = f"{base_name}_{suffix}"
    else:
        col_name = suffix
    if col_name in df.columns:
        ii = 0
        while True:
            test_col_name = f"{col_name}_{ii}"
            ii += 1
            if test_col_name not in df.columns:
                col_name = test_col_name
                break
    return col_name


def compress_mesh_data(mesh, cname="lz4"):
    if mesh.voxel_scaling is not None:
        vxsc = mesh.voxel_scaling
    else:
        vxsc = None
    mesh.voxel_scaling = None
    zvs = blosc.compress(mesh.vertices.tobytes(), typesize=8, cname=cname)
    zfs = blosc.compress(mesh.faces.tobytes(), typesize=8, cname=cname)
    zes = blosc.compress(mesh.link_edges.tobytes(), typesize=8, cname=cname)
    znm = blosc.compress(mesh.node_mask.tobytes(), typesize=1, cname=cname)
    mesh.voxel_scaling = vxsc
    return zvs, zfs, zes, znm, vxsc


def decompress_mesh_data(zvs, zfs, zes, znm, vxsc):
    vs = np.frombuffer(blosc.decompress(zvs), dtype=np.float).reshape(-1, 3)
    fs = np.frombuffer(blosc.decompress(zfs), dtype=np.int).reshape(-1, 3)
    es = np.frombuffer(blosc.decompress(zes), dtype=np.int).reshape(-1, 2)
    nm = np.frombuffer(blosc.decompress(znm), dtype=np.bool)
    return vs, fs, es, nm, vxsc


class MaskedMeshMemory(object):
    def __init__(self, mesh, index_only=False):
        self.node_mask = mesh.node_mask.copy()
        self.map_indices_to_unmasked = mesh.map_indices_to_unmasked
        self.map_boolean_to_unmasked = mesh.map_boolean_to_unmasked
        self.filter_unmasked_boolean = mesh.filter_unmasked_boolean
        self.filter_unmasked_indices = mesh.filter_unmasked_indices
        self.filter_unmasked_indices_padded = mesh.filter_unmasked_indices_padded
        self._voxel_scaling = mesh.voxel_scaling
        if index_only is False:
            self._kdtree = mesh.kdtree

    @property
    def kdtree(self):
        return self._kdtree

    @property
    def voxel_scaling(self):
        return self._voxel_scaling




from scipy import spatial, sparse
from meshparty import utils
import time
import numpy as np
from meshparty import trimesh_io
import logging
try:
    from caveclient import CAVEclient
except ImportError:
    logging.warning(
        "Need to pip install caveclient to repair mesh with pychunkedgraph")


def np_shared_rows(A, B):
    """ a utility to find intersection of two arrays

    Parameters
    ----------
    A: np.array
        a Nx2 array of np.int32 values
    B: np.array
        a Mx2 array of np.int32 values

    Returns
    -------
    np.array
        a Kx2 array of np.int32 values where the rows of A were also in B
    """
    A_flat = np.ascontiguousarray(np.sort(A, axis=1)).view(dtype=('i8,i8'))
    B_flat = np.ascontiguousarray(np.sort(B, axis=1)).view(dtype=('i8,i8'))
    good_edges = np.where(np.in1d(A_flat, B_flat))[0]
    return good_edges


def find_close_edges(vertices, labels, label_a, label_b):
    """ a function to find the closest vertex in one compoment for every vertex of another
    component

    Parameters
    ----------
    vertices: np.array
        a Nx3 array of mesh vertex locations
    labels: np.array
        A N np.array of ints across vertices where connected components have the same label
    label_a: int
        the label of the first connected component
    label_b: int
        the label of the second connected component

    Returns
    -------
    np.array
        close_edges, a Kx2 array (where K is the np.sum(labels==label_a) of vertex indices,
        where the first column entry is all the vertices in  component a
        and the second column entry is the index of the vertex in component b that is closest
        to the vertex in the first column.
    """
    a_inds = np.where(labels == label_a)[0]
    b_inds = np.where(labels == label_b)[0]
    b_tree = spatial.cKDTree(vertices[b_inds, :], balanced_tree=False)
    ds, closest = b_tree.query(vertices[a_inds, :], k=1)
    b_close = b_inds[closest[~np.isinf(ds)]]
    a_close = a_inds[~np.isinf(ds)]
    return np.hstack((a_close[:, np.newaxis], b_close[:, np.newaxis]))


def find_close_edges_sym(vertices, labels, label_a, label_b):
    """ a function to find the mutually closest vertices between two compoments of a mesh
    meaning.. vertex 1 is the closest vertex from component a to vertex 2
    and vertex 2 is the closest vertex in component b to vertex 1.

    Parameters
    ----------
    vertices: np.array
        a Nx3 array of mesh vertex locations
    labels: np.array
        A N np.array of ints across vertices where connected components have the same label
    label_a: int
        the label of the first connected component
    label_b: int
        the label of the second connected component

    Returns
    -------
    np.array
        close_edges, a Kx2 array of vertex indices,
        where the first column entry is an index in component a
        the second column entry is an index in component b
        and each vertex is mutually the closest vertex to the other component
    """

    new_edges_a = find_close_edges(vertices,
                                   labels,
                                   label_a,
                                   label_b)
    new_edges_b = find_close_edges(vertices,
                                   labels,
                                   label_b,
                                   label_a)
    is_both = np_shared_rows(new_edges_a, new_edges_b)
    return new_edges_a[is_both, :]


def find_all_close_edges(vertices, labels, ccs):
    """ Find all the mutually closest edges between all components of the mesh

    Parameters
    ----------
    vertices: np.array
        a Nx3 array of xyz vertex position
    labels: np.array
        a labelling of the vertices that reflect components of the mesh
    ccs: int
        the number of connected components in the mesh

    Returns
    -------
    np.array
        a Kx2 array of vertex indices that are mutually closest to each other
        across all combinations of components
    """
    all_edges = []
    for i in range(ccs):
        for j in range(i, ccs):
            all_edges.append(find_close_edges_sym(vertices, labels, i, j))
    return np.vstack(all_edges)


def find_edges_to_link(mesh, vert_ind_a, vert_ind_b, distance_upper_bound=2500, verbose=False):
    '''Given a mesh and two points on that mesh
    find a way to add edges to the  mesh graph so that those indices
    are on the same connected component

    Parameters
    ----------
    mesh: trimesh_io.Mesh
        a mesh to find edges on
    vert_ind_a: int
        one index into mesh.vertices, the first point
    vert_ind_b: int
        a second index into mesh.vertices, the second point
    distance_upper_bound: float
        a maximum distance to (default 2500 in units of mesh.vertices)
    verbose: bool
        whether to print debug info

    Returns
    -------
    np.array
        a Kx2 array of mesh indices that represent edges to add to the mesh to link the two points
        in a way that creates the shortest path between the points across mututally closest vertices
        from connected components.. not adding edges if they are larger than distance_upper_bound
        TODO: distance_upper_bound not presently implemented
    '''
    timings = {}
    start_time = time.time()

    # find the distance between the merge points and their center
    d = np.linalg.norm(
        mesh.vertices[vert_ind_a, :] - mesh.vertices[vert_ind_b, :])
    c = np.mean(mesh.vertices[[vert_ind_a, vert_ind_b], :], axis=0)
    # cut down the mesh to only include mesh vertices near the center of this
    # merge edge and within 2x the euclidean length of the edge
    inds = mesh.kdtree.query_ball_point(c, d*2)
    # convert this to a mask
    mask = np.zeros(len(mesh.vertices), dtype=np.bool)
    mask[inds] = True

    timings['create_mask'] = time.time()-start_time
    start_time = time.time()
    # create a masked version of the mesh
    mask_mesh = mesh.apply_mask(mask)

    timings['apply_mask'] = time.time()-start_time
    start_time = time.time()
    ccs, labels = sparse.csgraph.connected_components(
        mask_mesh.csgraph, return_labels=True)

    # map the original indices into this masked space
    mask_inds = mask_mesh.filter_unmasked_indices(
        np.array([vert_ind_a, vert_ind_b]))

    timings['masked_ccs'] = time.time()-start_time
    start_time = time.time()

    # find all the multually closest edges between the linked components
    new_edges = find_close_edges_sym(mask_mesh.vertices,
                                     labels,
                                     labels[mask_inds[0]],
                                     labels[mask_inds[1]])
    timings['find_close_edges_sym'] = time.time()-start_time
    start_time = time.time()

    # if there is now way to do this, fall back to adding all
    # edges that are close
    if len(new_edges) == 0:
        if verbose:
            print('finding all close edges')
        new_edges = find_all_close_edges(mask_mesh.vertices, labels, ccs)
        if verbose:
            print(f'new_edges shape {new_edges.shape}')
    # if there are still not edges we have a problem
    if len(new_edges) == 0:
        raise Exception('no close edges found')

    # create a new mesh that has these added edges
    # new_mesh = make_new_mesh_with_added_edges(mask_mesh, new_edges)
    total_edges = np.vstack([mask_mesh.graph_edges, new_edges])
    graph = utils.create_csgraph(mask_mesh.vertices, total_edges)
    timings['make_new_mesh'] = time.time()-start_time
    start_time = time.time()

    # find the shortest path to one of the linking spots in this new mesh
    d_ais_to_all, pred = sparse.csgraph.dijkstra(graph,
                                                 indices=mask_inds[0],
                                                 unweighted=False,
                                                 directed=False,
                                                 return_predecessors=True)
    timings['find_close_edges_sym'] = time.time()-start_time
    start_time = time.time()
    # make sure we found a good path
    if np.isinf(d_ais_to_all[mask_inds[1]]):
        raise Exception(
            f"cannot find link between {vert_ind_a} and {vert_ind_b}")

    # turn this path back into a original mesh index edge list
    path = utils.get_path(mask_inds[0], mask_inds[1], pred)
    path_as_edges = utils.paths_to_edges([path])
    good_edges = np_shared_rows(path_as_edges, new_edges)
    good_edges = np.sort(path_as_edges[good_edges], axis=1)
    timings['remap answers'] = time.time()-start_time
    if verbose:
        print(timings)
    return mask_mesh.map_indices_to_unmasked(good_edges)


def merge_points_to_merge_indices(mesh, merge_event_points, close_map_distance=300):
    """ function to take merge points in 3d space and find the right mesh.vertices indices
    filtering out merge events that merge the same component or are far away

    Parameters
    ----------
    mesh : trimesh_io.Mesh
        mesh to map points to
    merge_event_points: np.array
        a Nx2x3 matrix of merge points, where the last dimension is xyz
    close_map_distance: float or int
        a maximum distance to map points to indices (default 300)

    Returns
    -------
    np.array
        close_inds, a Mx2 matrix of indices into mesh.vertices
        M<N because some merge pairs might not have been able to both be mapped
        in under close_map_distance, and merge_points that only seem to connect
        already connected components are not added.
    """
    Nmerge = merge_event_points.shape[0]

    # find the connected components of the mesh
    ccs, labels = sparse.csgraph.connected_components(
        mesh.csgraph, return_labels=True)
    uniq_labels, label_counts = np.unique(labels, return_counts=True)
    large_cc_mask = np.isin(labels, uniq_labels[label_counts > 100])

    # convert the edge list to a Nmergex3 list of vertex positions
    merge_edge_verts = merge_event_points.reshape((merge_event_points.shape[0]*merge_event_points.shape[1],
                                                   merge_event_points.shape[2]))
    # create a fake list of indices for a skeleton
    # merge_edge_inds = np.arange(0, len(merge_edge_verts)).reshape(Nmerge, 2)

    # make a kdtree of the mesh triangle centers
    tree = spatial.cKDTree(mesh.triangles_center, balanced_tree=False)
    # find the close triangle centers to the merge vertices
    ds2, close_face_inds = tree.query(merge_edge_verts)
    # pick one of the triangle faces as our close index
    close_inds = mesh.faces[close_face_inds, 0]

    # lookup what connected component each close index is
    # reshaping it into a Nmerge x 2 format
    cc_labels = labels[close_inds].reshape((Nmerge, 2))

    # we want to differeniate merges that connected different
    # connected components from those that are connected disconnected
    # components of the mesh
    is_join_merge = cc_labels[:, 0] != cc_labels[:, 1]

    # however we can't trust that our mapping step worked perfectly
    # as euclidean distance can be tricky... so we do a second check
    # on the merge edges that end up on the same connected component
    # to see if there is another close vertex from a different connected
    # component

    # these are the merge edges that had 'close' results on at least one
    # of their edges
    close_mapped = np.min(ds2.reshape((Nmerge, 2)),
                          axis=1) < close_map_distance
    # calculate the index of merge edges that are mapped closely,
    # but ended up being on the same connected component
    # these are remapping candidates
    remap_merges = np.where(close_mapped & ~is_join_merge)[0]
    if (np.sum(remap_merges) == 0):
        return close_inds.reshape((Nmerge, 2))[is_join_merge, :]

    # pick out the coordinates of our candidates
    merge_coords = merge_event_points[close_mapped, :, :]
    # figure out which of the endpoint of the merge edge was the closest one
    close_merge = np.argmin(ds2.reshape((Nmerge, 2)), axis=1)

    # we want to reconsider the other end of the merge edge and get their xyz coords
    far_points = merge_event_points[remap_merges,
                                    (1-close_merge)[remap_merges], :]
    # If no candidates are available, return an empty array
    if len(far_points) == 0:
        return np.zeros((0, 2))

    # this is the connected component label for the close_merge end of the merge point
    closest_label = cc_labels[np.arange(0, len(cc_labels)), close_merge]
    # use a query_ball_point to find all the potential 'close' partners to the
    # far point
    close_points = tree.query_ball_point(far_points, close_map_distance)

    # iterate over the far points
    for k, (close_cc, arr) in enumerate(zip(closest_label[remap_merges], close_points)):
        # k is the index of the remap candidate
        # close_cc is the connected component of the other side of that edge
        # arr is the list of candidate triangle center points

        # calculate a binary vector over the candidate points as to whether it is part
        # of a different connected component than the other side of the merge edge
        other_comp_points = labels[mesh.faces[arr, 0]] != close_cc
        # find out how far each of the candidates that is on a different connected component is
        d = np.linalg.norm(
            far_points[k] - mesh.triangles_center[arr][other_comp_points], axis=1)

        # if this is empty that there are no candidates and we can safely ignore this merge edge
        if (d.shape[0] != 0):
            # otherwise there is another connected component that would be a 'close' match
            # this reindexs k to the 'far' point on the edge into the linear index of 'close_inds'
            ind = remap_merges[k]*2 + (1-close_merge[remap_merges[k]])
            # remap the index of the closest point with a different connected component
            # into the mesh vertex index.
            close_inds[ind] = mesh.faces[arr[np.where(
                other_comp_points)[0][np.argmin(d)]], 0]
            # reset the is_join_merge index to True for this case
            is_join_merge[remap_merges[k]] = True

    # return the close_inds reshaped into Nmerge x 2, and filtered to only be those
    # that are connected differente connected components
    return close_inds.reshape((Nmerge, 2))[is_join_merge, :]


def merge_log_to_points(merge_log, base_resolution):
    """Process the raw merge log into points in the euclidean space.
    """
    if isinstance(merge_log, dict):
        ml = merge_log['merge_edge_coords']
    else:
        ml = merge_log
    merge_event_points = np.array(ml).squeeze()

    if len(merge_event_points.shape) != 3:
        merge_event_points = merge_event_points.reshape(-1, 2, 3)

    return merge_event_points * base_resolution


def merge_log_edges(mesh, merge_log, base_resolution, close_map_distance=300, verbose=False):
    """ Process a merge log into mesh link edges
    """
    merge_event_points = merge_log_to_points(merge_log, base_resolution)

    # map these merge edge coordinates to indices on the mesh
    if len(merge_event_points) > 0:
        merge_edge_inds = merge_points_to_merge_indices(mesh,
                                                        merge_event_points,
                                                        close_map_distance=close_map_distance)
    else:
        merge_edge_inds = []

    if verbose:
        print(len(merge_edge_inds), len(merge_event_points))
    # iterate over these merge indices and find the minimal edge
    # that links these two connected componets
    total_link_edges = []

    for merge_ind in merge_edge_inds:
        link_edges = find_edges_to_link(mesh, merge_ind[0], merge_ind[1])
        total_link_edges.append(link_edges)

    # reshape the combined result into a M x 2 matrix
    if len(total_link_edges) > 0:
        link_edges = np.concatenate(total_link_edges)
    else:
        link_edges = np.array([])
    link_edges = link_edges.reshape((len(link_edges), 2))

    return link_edges


def get_link_edges(mesh, seg_id, datastack_name=None, close_map_distance=300,
                   server_address=None, verbose=False, client=None):
    """function to get a set of edges that should be added to a mesh

    Parameters
    ----------
    mesh : trimesh_io.Mesh
        the mesh to add edges to
    seg_id : np.uint64 or int
        the seg_id to query the PCG endpoint for merges
    dataset_name: str
        the name of the dataset to query
    close_map_distance: int or float
        the maximum distance to map (default 300 in units of mesh.vertices)
    server_address: str
        the url to the root of the CAVE deployment
    verbose: bool
        whether to print debug statements
    client : caveclient.ChunkedGraphClient

    Returns
    -------
    np.array
        link_edges, a Kx2 array of mesh.vertices indices representing edges that should be added to the mesh graph

    """

    # initialize a chunkedgraph client
    if client is None:
        client = CAVEclient(
            datastack_name, server_address=server_address).chunkedgraph

    # get the merge log
    if type(seg_id) is np.int64:
        seg_id = int(seg_id)

    merge_log = client.get_merge_log(seg_id)

    return merge_log_edges(mesh, merge_log, client.base_resolution,
                           close_map_distance=close_map_distance,
                           verbose=verbose)



import networkx as nx
from meshparty.meshwork import Meshwork
from meshparty.trimesh_io import Mesh
import pandas as pd
import numpy as np
from scipy import sparse


def _build_multicut_graph(nrn):
    G = nx.from_scipy_sparse_matrix(nrn.mesh.csgraph)

    G.add_node('source')
    G.add_node('target')

    source_edges = [('source', ii, np.inf)
                    for ii in nrn.anno['st_df'].df.query('type == "s"').mesh_index.values]
    target_edges = [('target', ii, np.inf)
                    for ii in nrn.anno['st_df'].df.query('type == "t"').mesh_index.values]

    G.add_weighted_edges_from(source_edges)
    G.add_weighted_edges_from(target_edges)

    return G


def _multicut_partitions(G, nrn):
    _, partition = nx.minimum_cut(G, 'source', 'target', capacity='weight')

    part0 = list(partition[0].difference({'source', 'target'}))
    part1 = list(partition[1].difference({'source', 'target'}))

    return nrn.MeshIndex(part0).to_mesh_mask_base, nrn.MeshIndex(part1).to_mesh_mask_base


def _build_nrn_with_st_annos(mesh, source_points, target_points):
    if isinstance(source_points, np.ndarray):
        source_points = source_points.tolist()
    if isinstance(target_points, np.ndarray):
        target_points = target_points.tolist()
    nrn = Meshwork(mesh, voxel_resolution=[1, 1, 1])
    source_df = pd.DataFrame(data={'pt_position': source_points})
    source_df['type'] = 's'
    target_df = pd.DataFrame(data={'pt_position': target_points})
    target_df['type'] = 't'
    st_df = source_df.append(target_df, ignore_index=True)
    st_df['pt_position'] = np.vstack(st_df['pt_position'].values).tolist()

    nrn.add_annotations('st_df', st_df, point_column='pt_position', anchored=True, overwrite=True)
    return nrn


def _build_local_mask(nrn, initial_window):
    ds = sparse.csgraph.dijkstra(
        nrn.mesh.csgraph, indices=nrn.anno['st_df'].mesh_index, limit=initial_window)
    d_sq = ds[:, nrn.anno['st_df'].mesh_index]
    if np.any(np.isinf(d_sq.ravel())):
        raise ValueError(
            "Initial window is too low (default: 10000) or points are in different components")

    # Centers mask on the point with the lowest mean distance to other points
    ctr_ind = np.argmin(np.mean(d_sq, axis=0))
    ctr_pt = nrn.anno['st_df'].mesh_index[ctr_ind]
    local_mask = np.invert(np.isinf(sparse.csgraph.dijkstra(
        nrn.mesh.csgraph, indices=ctr_pt, limit=np.max(d_sq[ctr_ind])+1)))
    return local_mask


def _faces_to_keep(p1mask, p2mask, nrn):
    p1mask_full = nrn.mesh.filter_unmasked_boolean(p1mask)
    p1faces = np.all(p1mask_full[nrn.mesh.faces], axis=1)

    p2mask_full = nrn.mesh.filter_unmasked_boolean(p2mask)
    p2faces = np.all(p2mask_full[nrn.mesh.faces], axis=1)

    neither_mask = np.invert(np.logical_or(p1mask_full, p2mask_full))
    nfaces = np.any(neither_mask[nrn.mesh.faces], axis=1)
    good_faces = np.logical_or(np.logical_or(p1faces, p2faces), nfaces)
    return good_faces


def _add_expected_edges(G, new_mesh, p1mask, p2mask, local_network_mask, test_split=True):
    "Adds edges that were not included in the faces graph"
    G.remove_node('source')
    G.remove_node('target')

    new_mesh_filt = new_mesh.apply_mask(local_network_mask)
    p1s = new_mesh_filt.filter_unmasked_boolean(p1mask)
    p2s = new_mesh_filt.filter_unmasked_boolean(p2mask)

    # Make matrix without cross-partition edges
    Gorig = nx.to_scipy_sparse_matrix(G)
    ii, jj, dd = sparse.find(Gorig)
    keep11 = p1s[ii] & p1s[jj]
    keep22 = p2s[ii] & p2s[jj]
    keep_all = keep11 | keep22

    GsplitB = sparse.csr_matrix((dd[keep_all], (ii[keep_all], jj[keep_all]))).toarray() > 0

    Gnew = new_mesh_filt.csgraph.toarray()
    GnewB = Gnew > 0

    # Places where edge in expected Gmat but not in new mesh
    link_edges_to_add_rough = np.vstack(np.where(np.logical_and(GsplitB == True, GnewB == False))).T
    if len(link_edges_to_add_rough) > 0:
        link_edges_to_add = np.unique(
            [tuple(x) for x in np.sort(link_edges_to_add_rough, axis=1)], axis=0)

        link_edges_unmasked = new_mesh_filt.map_indices_to_unmasked(link_edges_to_add)
        new_mesh.link_edges = np.vstack(
            (new_mesh.link_edges, new_mesh.filter_unmasked_indices(link_edges_unmasked)))

    if test_split:
        if len(link_edges_to_add_rough) > 0:
            new_mesh_filt.link_edges = np.vstack((new_mesh_filt.link_edges, link_edges_to_add))

        ncomp = sparse.csgraph.connected_components(new_mesh_filt.csgraph)[0]
        if ncomp > 2:
            print('Warning: more than 2 local components after split')
        if ncomp == 1:
            print('Warning: Only 1 local component after split')

    return new_mesh


def mesh_multicut(mesh, source_points, target_points, initial_window=10000, return_masks=False):
    """Use multi-point source/target split to cut a minimal set of faces from a mesh.
    Warns if the split produces more than 2 graph components in a local cutout, although
    the end result may still be suitable.

    Parameters
    ----------
    mesh : Mesh object
        Mesh to split
    source_points : np.array
        Nx3 numpy array of Euclidean points on one side of the desired partition
    target_points : np.array
        Mx3 numpy array of Euclidean points on the other side of the desired partition
    initial_window : int or float, optional
        Search window for a point-to-point distance matrix, needs to be larger than the max distance between source and/or target points along the mesh, by default 10000
    return_masks : bool, optional
        If True, returns vertex masks that denote each partition. Default is False

    Returns
    -------
    split_mesh : Mesh object
        Mesh with the same vertices as the original, but faces split per the multicut.
    partition_mask_source : np.array
        Boolean mask with True for mesh nodes in source partition. Returned if return_masks
        is True.
    partition_mask_target : np.array
        Boolean mask with True for mesh nodes in source partition. Returned if return_masks
        is True.
    """
    nrn = _build_nrn_with_st_annos(mesh, source_points, target_points)

    local_network_mask = _build_local_mask(nrn, initial_window)
    curr_mask = nrn.mesh_mask

    nrn.apply_mask(local_network_mask)

    G = _build_multicut_graph(nrn)
    p1mask, p2mask = _multicut_partitions(G, nrn)

    nrn.reset_mask()
    nrn.apply_mask(curr_mask)

    keep_faces = _faces_to_keep(p1mask, p2mask, nrn)

    new_mesh = Mesh(vertices=nrn.mesh.vertices,
                    faces=nrn.mesh.faces[keep_faces], node_mask=nrn.mesh_mask, link_edges=nrn.mesh.link_edges)

    new_mesh = _add_expected_edges(G, new_mesh, p1mask, p2mask, local_network_mask)

    if return_masks:
        return new_mesh, p1mask, p2mask
    else:
        return new_mesh



import numpy as np
from .utils import window_matrix
import numba

#####################
# Split by synapses #
#####################


def split_axon_by_synapses(
    nrn, pre_inds, post_inds, return_quality=True, extend_to_segment=True
):
    """Split a skeleton where the most paths flow between pre and postsynaptic synapse points.
    
    This method assumes that the axon can be split off by a single skeleton point.

    Parameters
    ----------
    sk : skeleton.Skeleton 
        Skeleton with N vertices to split
    pre_inds : array
        Indices of presynaptic sites
    post_inds : array
        Indices of postsynaptic sites
    return_quality : bool, optional
        If True (default), also returns the split quality
    extend_to_segment : bool, optional
        If True (default), places the cut at the proximal base of the segment where the max flow occurs.
        Probably does not make sense to use if the soma is not root.
    
    Returns
    -------
    is_axon : N array
        Boolean array that is True on the axon (presynaptic) side of the neuron.
    split_quality : float
        If return_quality is True, this is a number between 0 and 1 measuring how well the split
        segregates inputs from outputs. 1 is fully segregated, 0 is fully mixed.
    """
    if nrn.skeleton is None:
        raise ValueError("Meshwork must have skeleton")

    pre_inds = nrn._convert_to_meshindex(pre_inds)
    post_inds = nrn._convert_to_meshindex(post_inds)
    pre_inds = pre_inds[pre_inds.to_skel_index_padded >= 0]
    post_inds = post_inds[post_inds.to_skel_index_padded >= 0]

    axon_split = _find_axon_split(
        nrn.skeleton,
        pre_inds.to_skel_index_padded,
        post_inds.to_skel_index_padded,
        return_quality=return_quality,
        extend_to_segment=extend_to_segment,
    )

    if return_quality:
        axon_split_ind, split_quality = axon_split
    else:
        axon_split_ind = axon_split
    is_axon_sk = np.full(len(nrn.skeleton.vertices), False)
    is_axon_sk[nrn.skeleton.downstream_nodes(axon_split_ind)] = True
    is_axon = nrn.SkeletonIndex(np.flatnonzero(is_axon_sk)).to_mesh_index

    if return_quality:
        return is_axon, split_quality
    else:
        return is_axon


def axon_split_quality(is_axon, pre_inds, post_inds):
    axon_pre = sum(is_axon[pre_inds])
    axon_post = sum(is_axon[post_inds])
    dend_pre = sum(~is_axon[pre_inds])
    dend_post = sum(~is_axon[post_inds])

    counts = np.array([[axon_pre, axon_post], [dend_pre, dend_post]])
    observed_ent = _distribution_split_entropy(counts)

    unsplit_ent = _distribution_split_entropy([[len(pre_inds), len(post_inds)]])

    return 1 - observed_ent / unsplit_ent


def _distribution_split_entropy(counts):
    if np.sum(counts) == 0:
        return 0
    ps = np.divide(
        counts,
        np.sum(counts, axis=1)[:, np.newaxis],
        where=np.sum(counts, axis=1)[:, np.newaxis] > 0,
    )
    Hpart = np.sum(np.multiply(ps, np.log2(ps, where=ps > 0)), axis=1)
    Hws = np.sum(counts, axis=1) / np.sum(counts)
    Htot = -np.sum(Hpart * Hws)
    return Htot


def _precompute_synapse_inds(sk, syn_inds):
    Nsyn = len(syn_inds)
    n_syn = np.zeros(len(sk.vertices))
    for ind in syn_inds:
        n_syn[ind] += 1
    return Nsyn, n_syn


def _synapse_betweenness(sk, pre_inds, post_inds, use_entropy=False):

    Npre, n_pre = _precompute_synapse_inds(sk, pre_inds)
    Npost, n_post = _precompute_synapse_inds(sk, post_inds)

    syn_btwn = np.zeros(len(sk.vertices))
    split_index = np.zeros(len(sk.vertices))
    cov_paths_rev = sk.cover_paths[::-1]
    if use_entropy:
        entropy_normalization = distribution_split_entropy(
            np.array([[Npre, Npost], [0, 0]])
        )
    for path in cov_paths_rev:
        downstream_pre = 0
        downstream_post = 0
        for ind in path:
            downstream_pre += n_pre[ind]
            downstream_post += n_post[ind]
            syn_btwn[ind] = (
                downstream_pre * (Npost - downstream_post)
                + (Npre - downstream_pre) * downstream_post
            )
            if use_entropy:
                counts = np.array(
                    [
                        [downstream_pre, downstream_post],
                        [Npost - downstream_post, Npre - downstream_pre],
                    ]
                )
                split_index[ind] = (
                    1 - distribution_split_entropy(counts) / entropy_normalization
                )
        # Deposit each branch's synapses at the branch point.
        bp_ind = sk.parent_nodes(path[-1])
        if bp_ind is not None:
            n_pre[bp_ind] += downstream_pre
            n_post[bp_ind] += downstream_post
    if use_entropy:
        return syn_btwn, split_index
    else:
        return syn_btwn


def _find_axon_split(
    sk, pre_inds, post_inds, return_quality=True, extend_to_segment=True
):
    syn_btwn = _synapse_betweenness(sk, pre_inds, post_inds)
    high_vinds = np.flatnonzero(syn_btwn == max(syn_btwn))
    close_vind = high_vinds[np.argmin(sk.distance_to_root[high_vinds])]

    if return_quality:
        axon_qual_label = np.full(len(sk.vertices), False)
        axon_qual_label[sk.downstream_nodes(close_vind)] = True
        split_quality = axon_split_quality(axon_qual_label, pre_inds, post_inds)

    if extend_to_segment:
        relseg = sk.segment_map[close_vind]
        min_ind = np.argmin(sk.distance_to_root[sk.segments[relseg]])
        axon_split_ind = sk.segments[relseg][min_ind]
    else:
        axon_split_ind = close_vind

    if return_quality:
        return axon_split_ind, split_quality
    else:
        return axon_split_ind

        ############################
        # Topological branch order #
        ############################


def branch_order(nrn, return_as_skel=False):
    """Compute simple branch order, counting how many branch points are between a node and root.
    
    Parameters
    ----------
    nrn : Meshwork
        Meshwork with skeleton
    return_as_skel : bool, optional
        If True, returns an array for the skeleton instead of the mesh. Default is False.
    
    Returns
    -------
    branch_order
        N-length array containing the branch order for each mesh (or skeleton) vertex.
    """
    if nrn.skeleton is None:
        raise ValueError("Meshwork must have skeleton")

    has_bp = np.zeros(nrn.skeleton.n_vertices)
    has_bp[nrn.skeleton.branch_points] = 1

    covers = nrn.skeleton.cover_paths
    branch_order = np.zeros(nrn.skeleton.n_vertices)
    for cp in covers:
        path = cp[::-1]
        root_ind = nrn.skeleton.parent_nodes(path[0])
        if root_ind != -1:
            root_order = branch_order[root_ind]
        else:
            root_order = 0
        branch_order[path] = np.cumsum(has_bp[path]) + root_order
    branch_order[nrn.skeleton.root] = 0
    if return_as_skel:
        return branch_order
    else:
        return nrn.skeleton_property_to_mesh(branch_order, no_map_value=-1)


@numba.njit
def _strahler_path(baseline):
    out = np.full(len(baseline), -1, dtype=np.int64)
    last_val = 1
    for ii in np.arange(len(out)):
        if baseline[ii] > last_val:
            last_val = baseline[ii]
        elif baseline[ii] == last_val:
            last_val += 1
        out[ii] = last_val
    return out


def strahler_order(nrn, return_as_skel=False):
    """Computes strahler number. Tips have strahler number 1, and strahler number increments
    when a two branches with the same strahler number merge.
    
    Parameters
    ----------
    nrn : Meshwork
        Meshwork neuron with skeleton
    return_as_skel : bool, optional
        Description
    
    Returns
    -------
    strahler_number : array
        Array of mesh (or, optionally, skeleton) vertices with strahler number of vertex.
    """
    if nrn.skeleton is None:
        raise ValueError("Meshwork must have skeleton")
    covers = nrn.skeleton.cover_paths
    strahler = np.full(nrn.skeleton.n_vertices, -1)
    for cp in covers[::-1]:
        new_vals = _strahler_path(strahler[cp])
        strahler[cp] = new_vals
        pind = nrn.skeleton.parent_nodes(cp[-1])
        if pind >= 0:
            if strahler[cp[-1]] > strahler[pind]:
                strahler[pind] = strahler[cp[-1]]
            elif strahler[cp[-1]] == strahler[pind]:
                strahler[pind] += 1
    if return_as_skel:
        return strahler
    else:
        return nrn.skeleton_property_to_mesh(strahler, no_map_value=-1)

        ######################
        # Density estimation #
        ######################

    def _gaussian_kernel(x, sigma):
        def func(x):
            return np.exp(-x * x / (2 * sigma * sigma))

        return func

    def _normalize_flat(W, nrn, exclude_root):
        if exclude_root:
            g = nrn.skeleton.cut_graph(
                nrn.skeleton.child_nodes([nrn.skeleton.root])[0], directed=False
            )
            len_per = np.array(g.sum(axis=1) / 2).ravel()
        else:
            len_per = np.array(nrn.skeleton.csgraph_undirected.sum(axis=1) / 2).ravel()
        return W.dot(len_per.reshape(-1, 1)).ravel()

    def _normalize_gausssian(W):
        return W.sum(axis=1).squeeze()

    def linear_density(
        nrn,
        inds,
        width,
        weight=None,
        kernel="flat",
        normalize=True,
        exclude_root=False,
    ):
        """Compute a sliding window average linear density of points across the object
        
        Parameters
        ----------
        inds : array
            Mesh indices for density (e.g. synapse locations).
        width : numeric
            width of average window (in all directions).
        weight : array None, optional
            Weight for each point for weighted average. If None, assumes weight of unity.
        normalize : bool, optional
            If False, sums the weights but does not normalize by amount of cable.
            Default is True.
        exclude_root : bool, optional
            If True, disconnects root from the graph for the case that the root is not
            well-approximated by a line (e.g. a cell body.). The density for those vertices
            will be infinite or nan.
            Default is False.


        Returns
        -------
        density_estimate : array
            N-length array of density at all mesh vertices.
        """
        if kernel == "flat":
            W = window_matrix(nrn.skeleton, width)
        elif kernel == "gaussian":
            dist_func = _gaussian_kernel(width)
            W = window_matrix(nrn.skeleton, 4 * width, dist_func)
        inds = nrn._convert_to_meshindex(inds)
        has_inds = np.full(nrn.skeleton.n_vertices, 0)
        if weight is None:
            skinds, count = np.unique(inds.to_skel_index, return_counts=True)
            has_inds[skinds] = count
        else:
            for w, skind in zip(weight, inds.to_skel_index):
                has_inds[skind] += w
        item_count = W.dot(has_inds.reshape(-1, 1)).ravel()
        if normalize:
            if kernel == "flat":
                norm = _normalize_flat(W, nrn, exclude_root)
            elif kernel == "gaussian":
                norm = _normalize_gausssian(W)

            with np.errstate(divide="ignore"):
                rho = item_count / norm
        else:
            rho = item_count
        return rho[nrn.skeleton.mesh_to_skel_map][nrn.mesh.node_mask]



import trimesh.ray
from trimesh.ray import ray_pyembree
from trimesh.ray import ray_pyembree
from scipy.linalg import block_diag
import numpy as np
import multiwrapper.multiprocessing_utils as mu
from meshparty import trimesh_io
import logging


def ray_trace_distance(vertex_inds, mesh, max_iter=10, rand_jitter=0.001, verbose=False, ray_inter=None):
    '''
    Compute distance to opposite side of the mesh for specified vertex indices on the mesh.

    Parameters
    ----------
    vertex_inds : np.array
        a K long set of indices into the mesh.vertices that you want to perform ray tracing on
    mesh : :obj:`meshparty.trimesh_io.Mesh`
        mesh to perform ray tracing on
    max_iter : int
        maximum retries to attempt in order to get a proper sdf measure (default 10)
    rand_jitter : float
        the amplitude of gaussian jitter on the vertex normal to add on each iteration (default .001)
    verbose : bool
        whether to print debug statements (default False)
    ray_inter: ray_pyembree.RayMeshIntersector
        a ray intercept object pre-initialized with a mesh, in case y ou are doing this many times
        and want to avoid paying initialization costs. (default None) will initialize it for you

    Returns
    -------
    np.array
        rs, a K long array of sdf values. rays with no result after max_iters will contain zeros.

    '''
    if not trimesh.ray.has_embree:
        logging.warning(
            "calculating rays without pyembree, conda install pyembree for large speedup")

    if ray_inter is None:
        ray_inter = ray_pyembree.RayMeshIntersector(mesh)

    rs = np.zeros(len(vertex_inds))
    good_rs = np.full(len(rs), False)

    it = 0
    while not np.all(good_rs):
        if verbose:
            print(np.sum(~good_rs))
        blank_inds = np.where(~good_rs)[0]
        starts = (mesh.vertices -
                  mesh.vertex_normals)[vertex_inds, :][~good_rs, :]
        vs = -mesh.vertex_normals[vertex_inds, :] \
            + (1.2**it)*rand_jitter*np.random.rand(*
                                                   mesh.vertex_normals[vertex_inds, :].shape)
        vs = vs[~good_rs, :]

        rtrace = ray_inter.intersects_location(starts, vs, multiple_hits=False)

        if len(rtrace[0] > 0):
            # radius values
            rs[blank_inds[rtrace[1]]] = np.linalg.norm(
                mesh.vertices[vertex_inds, :][rtrace[1]]-rtrace[0], axis=1)
            good_rs[blank_inds[rtrace[1]]] = True
        it += 1
        if it > max_iter:
            break
    return rs


def vogel_disk_sampler(num_points, radius=1):
    """Uniform sampler of the unit disk.
    """
    golden_angle = np.pi * (3 - np.sqrt(5))
    thetas = golden_angle * np.arange(num_points)
    radii = radius * np.sqrt(np.arange(num_points)/num_points)

    xs = radii * np.cos(thetas)
    ys = radii * np.sin(thetas)

    return xs, ys


def unit_vector_sampler(num_points, widest_angle=np.pi/3):
    if np.abs(widest_angle) > np.pi:
        print('')
        return None
    xs, ys = vogel_disk_sampler(num_points, radius=1)
    zs = 1/np.tan(widest_angle) * np.ones(num_points)
    Vs = np.vstack((xs, ys, zs)).T
    return Vs / np.linalg.norm(Vs, axis=1)[:, np.newaxis]


def Rx(ang):
    return np.array([[1, 0, 0],
                     [0, np.cos(ang), -np.sin(ang)],
                     [0, np.sin(ang), np.cos(ang)]])


def Ry(ang):
    return np.array([[np.cos(ang), 0, np.sin(ang)],
                     [0,           1,       0],
                     [-np.sin(ang), 0, np.cos(ang)]])


def Rz(ang):
    return np.array([[np.cos(ang), -np.sin(ang), 0],
                     [np.sin(ang), np.cos(ang), 0],
                     [0,  0, 1]])


def _rotated_cone(data):
    phi, theta, vs_raw = data
    Rtrans = np.dot(Rz(phi), Ry(theta))
    return np.dot(Rtrans, vs_raw.T).T


def oriented_vector_cones(center_vectors, num_points, widest_angle=np.pi/3, normalize=False):
    """Produces all ray cones
    """
    if normalize:
        cv_norm = center_vectors / \
            np.linalg.norm(center_vector, axis=1)[:, np.newaxis]
    else:
        cv_norm = center_vectors

    thetas = np.arccos(cv_norm[:, 2])
    phis = np.arctan2(cv_norm[:, 1], cv_norm[:, 0])

    vs_raw = unit_vector_sampler(num_points, widest_angle=widest_angle)

    Rtranses = []
    data = []
    for phi, theta in zip(phis, thetas):
        data.append((phi, theta, vs_raw))
    vector_cones = mu.multiprocess_func(_rotated_cone, data)
    return vector_cones


def _multi_angle_weighted_distance(data):
    ds, angles, weights = data
    return angle_weighted_distance(ds, angles, weights)


def angle_weighted_distance(ds, angles, weights):
    """Does angle-weighted distance averaging. Ignores outliers and emphasizes normal hits.
    """
    if len(ds) == 0 or np.nansum(weights) == 0:
        return np.nan

    med_angle = np.median(angles)
    std_angle = np.std(angles)
    min_angle = med_angle-std_angle
    max_angle = max(med_angle+std_angle, np.pi / 2)
    good_rows = np.logical_and(angles >= min_angle, angles <= max_angle)

    nanaverage = np.nansum(
        ds[good_rows] * weights[good_rows]) / np.nansum(weights[good_rows])
    return nanaverage


def all_angle_weighted_distances(ds, angles, weights, rep_inds, inds):

    data = []
    real_inds, slice_bnds = np.unique(rep_inds, return_index=True)

    ind_map = []
    for ii in range(len(real_inds)-1):
        row = slice(slice_bnds[ii], slice_bnds[ii+1])
        data.append((ds[row], angles[row], weights[row]))
        ind_map.append(real_inds[ii])
    row = slice(slice_bnds[-1], len(ds))
    data.append((ds[row], angles[row], weights[row]))
    ind_map.append(real_inds[-1])

    rs = mu.multiprocess_func(_multi_angle_weighted_distance, data)

    rs_out = np.nan * np.zeros(len(inds))
    rs_out[np.array(ind_map)] = rs

    return rs_out


def _compute_ray_vectors(mesh, mesh_inds, num_points, cone_angle):
    return np.vstack(oriented_vector_cones(-mesh.vertex_normals[mesh_inds], num_points, cone_angle))


def shape_diameter_function(mesh_inds, mesh, num_points=30, cone_angle=np.pi/3):
    """Computes shape diameter function by sending a cone of rays from each specified vertex point
    and doing a weighted average of where they hit the opposite side of the mesh.

    Parameters
    ----------
    mesh : trimesh.Mesh
        Mesh 
    mesh_inds : list of indices
        Vertex indices at which to compute SDF
    num_points : int, optional
        Number of points per cones (default is 30)
    cone_angle : float, optional
        Angular width of the cone

    Returns
    -------

        Description
    """
    start = (mesh.vertices-mesh.vertex_normals)[mesh_inds, :]
    rep_inds = np.concatenate([ii*np.ones(num_points, dtype=int)
                               for ii in range(start.shape[0])])
    starts = start[rep_inds]

    vs = _compute_ray_vectors(mesh, mesh_inds, num_points, cone_angle)

    ray_inter = ray_pyembree.RayMeshIntersector(mesh)
    rtrace = ray_inter.intersects_location(starts, vs, multiple_hits=False)

    hit_rows = rtrace[1]
    ds = np.linalg.norm(rtrace[0] - starts[hit_rows], axis=1)
    angles = np.arccos(
        np.sum(mesh.face_normals[rtrace[2]] * vs[hit_rows], axis=1))
    with np.errstate(divide='ignore', invalid='ignore'):
        weights = 1/angles
    good_rows = np.isfinite(weights)

    rs = all_angle_weighted_distances(
        ds[good_rows], angles[good_rows], weights[good_rows], rep_inds[hit_rows[good_rows]], mesh_inds)
    return rs



import os
import h5py
import orjson
import json

from .utils_io import NumpyEncoder

import numpy as np
from meshparty import skeleton
from dataclasses import asdict

FILE_VERSION = 2


def write_skeleton_h5(sk, filename, overwrite=False):
    """
    Write a skeleton and its properties to an hdf5 file.

    Parameters
    ----------
    sk : :obj:`meshparty.skeleton.Skeleton`
        new_mesh_filt
    filename : str
        Filename of skeleton file.
    overwrite : bool
         Allows overwriting.(default False).
    """
    if not hasattr(sk, "meta"):
        sk.meta = {}

    write_skeleton_h5_by_part(
        filename,
        vertices=sk.vertices,
        edges=sk.edges,
        meta=sk.meta,
        mesh_to_skel_map=sk.mesh_to_skel_map,
        vertex_properties=sk.vertex_properties,
        root=sk.root,
        overwrite=overwrite,
    )


def write_skeleton_h5_by_part(
    filename,
    vertices,
    edges,
    meta,
    mesh_to_skel_map=None,
    vertex_properties={},
    root=None,
    overwrite=False,
):
    """
    Helper function for writing all parts of a skeleton file to an h5.

    Parameters
    ----------
    filename : str
        path to write
    vertices : np.array
        Nx3 numpy array of vertex locations
    edges : np.array
        Kx2 numpy array of vertex indices for edges
    mesh_to_skel_map : np.array
        M long numpy array.  M is the number of vertices in a mesh that this
        is associated with.  The entries are indices into the N skeleton vertices
    vertex_properties : dict
        a dictionary of np.arrays, were keys are descriptive labels
        and the values are arrays that quantify that label at each vertex
        examples..
        mesh_index) the mesh index of this vertex
        rs) the sdf (local caliber/thickness) of the mesh at each index
    root : int
        which vertex index is root
    overwrite : bool
        whether to overwrite file

    """

    if os.path.isfile(filename):
        if overwrite:
            os.remove(filename)
        else:
            return
    with h5py.File(filename, "w") as f:
        f.attrs["file_version"] = FILE_VERSION

        f.create_dataset("vertices", data=vertices, compression="gzip")
        f.create_dataset("edges", data=edges, compression="gzip")
        f.create_dataset(
            "meta",
            data=np.string_(
                orjson.dumps(asdict(meta), option=orjson.OPT_SERIALIZE_NUMPY)
            ),
        )
        if mesh_to_skel_map is not None:
            f.create_dataset(
                "mesh_to_skel_map", data=mesh_to_skel_map, compression="gzip"
            )
        if len(vertex_properties) > 0:
            _write_dict_to_group(f, "vertex_properties", vertex_properties)
        if root is not None:
            f.create_dataset("root", data=root)


def _write_dict_to_group(f, group_name, data_dict):
    d_grp = f.create_group(group_name)
    for d_name, d_data in data_dict.items():
        d_grp.create_dataset(d_name, data=json.dumps(d_data, cls=NumpyEncoder))


def read_skeleton_h5_by_part(filename):
    """
    Helper function for extracting all parts of a skeleton file from an h5.

    Parameters
    ----------
    filename : str
        path to a h5 file with skeletons

    Returns
    -------
    str
        filename, path to write
    np.array
        vertices, Nx3 numpy array of vertex locations
    np.array
        edges , Kx2 numpy array of vertex indices for edges
    np.array
        mesh_to_skel_map , M long numpy array.  M is the number of vertices in a mesh that this
        is associated with.  The entries are indices into the N skeleton vertices
    dict
        vertex_properties, a dictionary of np.arrays, were keys are descriptive labels
        and the values are arrays that quantify that label at each vertex
        examples..
        mesh_index) the mesh index of this vertex
        rs) the sdf (local caliber/thickness) of the mesh at each index
    int
        root, which vertex index is root
    bool
        overwrite, whether to overwrite file

    """
    assert os.path.isfile(filename)

    with h5py.File(filename, "r") as f:
        vertices = f["vertices"][()]
        edges = f["edges"][()]

        if "mesh_to_skel_map" in f.keys():
            mesh_to_skel_map = f["mesh_to_skel_map"][()]
        else:
            mesh_to_skel_map = None

        vertex_properties = {}
        if "vertex_properties" in f.keys():
            for vp_key in f["vertex_properties"].keys():
                vertex_properties[vp_key] = json.loads(
                    f["vertex_properties"][vp_key][()], object_hook=_convert_keys_to_int
                )

        if "meta" in f.keys():
            dat = f["meta"][()].tobytes()
            meta = orjson.loads(dat)
        else:
            meta = {}

        if "root" in f.keys():
            root = f["root"][()]
        else:
            root = None

    return vertices, edges, meta, mesh_to_skel_map, vertex_properties, root


def read_skeleton_h5(filename, remove_zero_length_edges=False):
    """
    Reads a skeleton and its properties from an hdf5 file.

    Parameters
    ----------
    filename: str
        path to skeleton file
    remove_zero_length_edges: bool, optional
        If True, post-processes the skeleton data to removes any zero
        length edges. Default is False.
    Returns
    -------
    :obj:`meshparty.skeleton.Skeleton`
        skeleton object loaded from the h5 file

    """
    (
        vertices,
        edges,
        meta,
        mesh_to_skel_map,
        vertex_properties,
        root,
    ) = read_skeleton_h5_by_part(filename)
    return skeleton.Skeleton(
        vertices=vertices,
        edges=edges,
        mesh_to_skel_map=mesh_to_skel_map,
        vertex_properties=vertex_properties,
        root=root,
        remove_zero_length_edges=remove_zero_length_edges,
        meta=meta,
    )


def swc_node_labels(
    sk,
    dendrite_indices=None,
    apical_indices=None,
    soma_indices=None,
    axon_indices=None,
    dendrite_default=True,
):
    """Assemble swc node labels based on compartment labels. By default, unlabeled indices are considered dendrite.

    Parameters
    ----------
    sk : Skeleton
        Neuron skeleton object with N vertices
    dendrite_indices : array, optional
        Array of indices associated with the dendrites (or basal dendrites if apicals are distinct), by default None.
    apical_indices : array, optional
        Array of indices (or boolean mask) for the apical dendrite, by default None.
    soma_indices : array, optional
        Array of indices (or boolean mask) for the soma, by default None.
    axon_indices : axon, optional
        Array of indices (or boolean mask) for the axon, by default None.
    dendrite_default : bool, optional,
        If True, assumed unlabeled vertices are dendrite. Otherwise, give a label of 0.

    Returns
    -------
    nodelabels
        N-length vector with the appropriate SWC label for each compartment. Unlabeled vertices are given a default label.
        Default label is 3 (basal dendrite) if dendrite default is True, else 0.
    """
    SOMA_LABEL = 1
    AXON_LABEL = 2
    DENDRITE_LABEL = 3
    APICAL_LABEL = 4

    inds = [dendrite_indices, apical_indices, soma_indices, axon_indices]
    labels = [DENDRITE_LABEL, APICAL_LABEL, SOMA_LABEL, AXON_LABEL]

    if dendrite_default:
        val = DENDRITE_LABEL
    else:
        val = 0
    node_labels = np.full(len(sk.vertices), val)
    for ii, label in zip(inds, labels):
        if ii is not None:
            node_labels[np.array(ii)] = label
    return node_labels


def export_to_swc(
    skel,
    filename,
    node_labels=None,
    radius=None,
    header=None,
    xyz_scaling=1000,
    resample_spacing=None,
    interp_kind="linear",
    tip_length_ratio=0.5,
):
    """
    Export a skeleton file to an swc file
    (see http://www.neuronland.org/NLMorphologyConverter/MorphologyFormats/SWC/Spec.html for swc definition)

    Parameters
    ----------
    filename : str
        path to the file to save the swc to
    node_labels : iterable
        None (default) or an interable of ints co-indexed with vertices.
        Corresponds to the swc node categories. Defaults to setting all
        nodes to label 3, dendrite.
    radius : iterable
        None (default) or an iterable of floats. This should be co-indexed with vertices.
        Radius values are assumed to be in the same units as the node vertices.
    header : str or list, default None.
        An optional header string for the file. Each element of the list
    xyz_scaling: Number, default 1000. Down-scales spatial units from the skeleton's units to
                        whatever is desired by the swc. E.g. nm to microns has scaling=1000.
    """

    if header is None:
        header_string = ""
    else:
        if isinstance(header, str):
            header = [header]
        header[0] = " ".join(["#", header[0]])
        header_string = "\n# ".join(header)

    if radius is None:
        radius = np.full(len(skel.vertices), xyz_scaling)
    elif np.issubdtype(type(radius), int):
        radius = np.full(len(skel.vertices), radius)

    if node_labels is None:
        node_labels = np.full(len(skel.vertices), 0)

    if resample_spacing is not None:
        skel, output_map = skeleton.resample(
            skel,
            spacing=resample_spacing,
            tip_length_ratio=tip_length_ratio,
            kind=interp_kind,
        )
        node_labels = node_labels[output_map]
        radius = radius[output_map]

    swc_dat = _build_swc_array(skel, node_labels, radius, xyz_scaling)

    np.savetxt(
        filename,
        swc_dat,
        delimiter=" ",
        header=header_string,
        comments="#",
        fmt=["%i", "%i", "%.3f", "%.3f", "%.3f", "%.3f", "%i"],
    )


def _build_swc_array(skel, node_labels, radius, xyz_scaling):
    """
    Helper function for producing the numpy table for an swc.
    """
    order_old = np.concatenate([p[::-1] for p in skel.cover_paths])
    new_ids = np.arange(skel.n_vertices)
    order_map = dict(zip(order_old, new_ids))

    node_labels = np.array(node_labels)[order_old]
    xyz = skel.vertices[order_old]
    radius = radius[order_old]
    par_ids = np.array([order_map.get(nid, -1) for nid in skel.parent_nodes(order_old)])

    swc_dat = np.hstack(
        (
            new_ids[:, np.newaxis],
            node_labels[:, np.newaxis],
            xyz / xyz_scaling,
            radius[:, np.newaxis] / xyz_scaling,
            par_ids[:, np.newaxis],
        )
    )
    return swc_dat


def _convert_keys_to_int(x):
    if type(x) is dict:
        return {int(k): v for k, v in x.items()}
    else:
        return x



import os


HOME = os.path.expanduser("~")
PACKAGE_PATH = "%s/../" % os.path.dirname(os.path.abspath(__file__))
EXAMPLE_PATH = "%s/example/" % PACKAGE_PATH

if not os.path.exists(EXAMPLE_PATH):
    os.makedirs(EXAMPLE_PATH)


def save_local_views(mesh, n_points, n_samples, save_name, pc_align=True):
    save_folder = "%s/%s/" % (EXAMPLE_PATH, save_name)

    if not os.path.exists(save_folder):
        os.makedirs(save_folder)

    for i_sample in range(n_samples):
        local_mesh = mesh.get_local_mesh(n_points=n_points,
                                         pc_align=pc_align)

        save_path = "%s/sample%d.obj" % (save_folder, i_sample)

        local_mesh.write_to_file(save_path)



if __name__ == "__main__":
    from meshparty import trimesh_io
    mm = trimesh_io.MeshMeta()
    mesh = mm.mesh("%s/MeshParty/example/3205058_m.obj" % HOME)

    save_local_views(mesh, n_points=2000, n_samples=10,
                     save_name="3205058_m_local_500")



from setuptools import setup
import re
import os
import codecs
from setuptools.command.test import test as TestCommand
import sys


class PyTest(TestCommand):
    user_options = [("pytest-args=", "a", "Arguments to pass to pytest")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ""

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import shlex
        import pytest

        self.pytest_args += (
            " --cov=meshparty --cov-report html --junitxml=test-reports/test.xml"
        )

        errno = pytest.main(shlex.split(self.pytest_args))
        sys.exit(errno)


with open("test_requirements.txt", "r") as f:
    test_required = f.read().splitlines()

with open("requirements.txt", "r") as f:
    required = f.read().splitlines()
    print(required)

dependency_links = []
del_ls = []
for i_l in range(len(required)):
    l = required[i_l]
    if l.startswith("-e"):
        dependency_links.append(l.split("-e ")[-1])
        del_ls.append(i_l)

        required.append(l.split("=")[-1])

for i_l in del_ls[::-1]:
    del required[i_l]

setup(
    use_scm_version=True,
    name="meshparty",
    description="a service to work with meshes",
    author="Sven Dorkenwald, Forrest Collman, Casey Schneider-Mizell",
    author_email="svenmd@princeton.edu, forrestc@alleninstitute.org, caseys@alleninstitute.org",
    url="https://github.com/sdorkenw/MeshParty.git",
    packages=["meshparty"],
    include_package_data=True,
    install_requires=required,
    extras_require={"SDF": ["pyembree"], "REPAIR": ["caveclient>=4.0.0"]},
    setup_requires=["pytest-runner", "setuptools_scm"],
    dependency_links=dependency_links,
    tests_require=test_required,
    cmdclass={"test": PyTest},
)
